/* Program to calculate values of Ao, Ac, PSo, PSc to be used in the
   ED model */

/* ****************************************************************** */
/* This version generates a lookup table indexed by light, CO2, temp, */
/* Vm0, and dewpoint. There is no reference to time or site.          */
/* ****************************************************************** */

/* handle files over 2GB */
#ifndef _LARGEFILE_SOURCE
# define _LARGEFILE_SOURCE
#endif
#define _FILE_OFFSET_BITS 64

#include "mech_lut.h"
#include "EDMech.h"

#include "farquar.h"
#include <getopt.h>
#include <stdlib.h>
#include <string.h>
#include <libiberty.h>
#include <assert.h>
#include <float.h>		/* provides FLT_EPSILON */

/* commandline option flags */
static int use_c3, use_c4;
int C_MIN = 0;
int C_MAX = 1;
#define CO2_DEFAULT 367		/* default concentration in ppm */
/** atmospheric CO2 concentration */
static unsigned int co2_ppm = CO2_DEFAULT;
static int co2_from_commandline = 0;  
static unsigned int year = 0;	/* only affects filenames */
static char binary_outfile_name[256];
# warning "hardcoded ascii output filename - should get from cmdline."
static char *ascii_outfile_name = "mechlut.txt";
static FILE *outfile_ascii = NULL;
static FILE *outfile_binary = NULL;

/* switches for output type */
static int BINARY_OUTPUT;
static int ASCII_OUTPUT;

/** Rn above which open case calculations do not fail. */
static const double Rn_EPSILON = 3.0;

struct solution soln;	/* Farquar solution output*/

static void printusage() {
  printf("Usage: EDMech [options] <output filename base>\n");
  printf("Options:\n");
  printf(" -a \tascii output (default off)\n");
  printf(" -b \tbinary output (default on)\n");
  printf(" -c <3|4>\tOutput only for c3 or c4 (default c3)\n");
  printf(" -C <int>\tCO2 concentration in PPM"
	 " (default to hard-coded range)\n");
}

/**
 * Parse argument list and set various flags.  Should also set I/O
 * filenames. Right now it returns the index of the first non-flag arg
 * instead.
 */
static int handle_args(int argc, char* argv[]) {
  int flag;
  extern int optind;
  use_c3 = 1;  use_c4 = 0;	/* these can be overwritten by the -c
				   flag */
  
  while((flag = getopt(argc, argv, "abc:C:y:")) != -1) {
    switch(flag){
    case 'a':
      ASCII_OUTPUT = 1;
      break;
    case 'b':
      BINARY_OUTPUT = 1;
      break;
    case 'c':
      if(strcmp(optarg, "3") == 0) {
	use_c4 = 0; use_c3 = 1;
      }
      else if(strcmp(optarg, "4") == 0) {
	use_c3 = 0; use_c4 = 1;
      } 
      else {
	fprintf(stderr,"bad arg to -c\n");
	printusage();
	exit(EXIT_FAILURE);
      }
      break;
    case 'C':
      co2_ppm = atoi(optarg);
      co2_from_commandline = 1;
      break;
    case 'y':			/* year only affects filenames */
      year = atoi(optarg);
      break;
    case '?':			/* any invalid arg */
      printusage();
      exit(1);
      break;
    } /* switch */
  } /* while */
    
  /* set defaults for things that didn't get set */
  if(use_c3 == 0 && use_c4 == 0) {
    use_c3 = 1; use_c4 = 1;
  }
  
  if(BINARY_OUTPUT == 0 && ASCII_OUTPUT == 0)
    BINARY_OUTPUT = 1;
    
  if(!use_c3) C_MIN = 1;
  if(!use_c4) C_MAX = 0;
    
  return(optind);
} /* handle_args */

int main(int argc, char *argv[])
{
  //  char * basename;		/* base for output filenames */
  int arg_index;		/* where non-flag args begin */
  
  //char filename_in[256];
  FILE *in_file = NULL;

  struct envPars theEnv;
  struct leafPars c3lf, c4lf;	/* c3 and c4 leaves */
  struct leafVars cl;		/* current leaf */
  struct pars cp;		/* current pars */

  /* Increments of LAI to get a final LAI above of 12 */
  /* may be used by call to init_param */
  const double Vm0_MIN = 6.25;
  const double Vm0_MAX = 18.75;

  /* lookup tables for binned values */
  struct param *Vm0 = NULL;
  struct param *CO2 = NULL;
  struct param *humidity = NULL;
  struct param *Rn = NULL;
  struct param *T = NULL;

  /* bin indexes */
  int CO2_index;
  int Vm0_index;
  int T_index;
  int humidity_index;
  int Rn_index;

  /* solver related variables */
  int status;			/* return flag for solver */
  int Cx;			/* 0 = C3 1 = C4 */
  int i;

  printf("epsilon = %e\n", EPS);

#if 1
  /* init luts */
  Vm0 = init_param("Vm0", N_Vm0, Vm0_MIN,
		 (Vm0_MAX - Vm0_MIN)/(N_Vm0 -1), LINEAR, "dimensionless");
#else
#warning "TEMPORARY Vm0 range"
  Vm0 = init_param("Vm0", 3, 12.5, 6.25, LINEAR, "dimensionless");
#endif

  /* param format:
     name, N, min, step, log/linear, units */

#if 0
  /* one of everything */
  CO2 = init_param("CO2", 1, 270, 5, LINEAR, "ppm");
  T = init_param("T", 1, -9, 1, LINEAR, "C");
  humidity = init_param("humidity", 1, 20, 1, LINEAR, "%");
  /* range 0 - ~900 */
  Rn = init_param("Rn", 1, 0, log10(900)/(17-1), LOGSCALE, "W/m^2");
#elif 0
  /* 5ppm co2 */
  CO2 = init_param("CO2", 89, 270, 5, LINEAR, "ppm");
  T = init_param("T", 50, -9, 1, LINEAR, "C");
  humidity = init_param("humidity", 80, 20, 1, LINEAR, "%");
  /* range 0 - ~900 */
  Rn = init_param("Rn", 17, 0, log10(900)/(17-1), LOGSCALE, "W/m^2");
#elif 1
  /* 1 co2 level */
#warning "1 CO2 level: 270"
  CO2 = init_param("CO2", 1, 270, 5, LINEAR, "ppm");
  T = init_param("T", 50, -9, 1, LINEAR, "C");
  humidity = init_param("humidity", 80, 20, 1, LINEAR, "%");
  /* range 0 - ~900 */
  Rn = init_param("Rn", 17, 0, log10(900)/(17-1), LOGSCALE, "W/m^2");
#elif 0
  /* maybe-realistic ranges */
  CO2 = init_param("CO2", 15, 270, 30, LINEAR, "ppm");
  T = init_param("T", 50, -9, 1, LINEAR, "C");
  humidity = init_param("humidity", 80, 20, 1, LINEAR, "%");
  /* range 0 - ~900 */
  Rn = init_param("Rn", 17, 0, log10(900)/(17-1), LOGSCALE, "W/m^2");
#else
  /* ranges for debugging */
  CO2 = init_param("CO2", 6, 220, 50, LINEAR, "ppm");
  T = init_param("T", 5, 1, 8, LINEAR, "C");
  humidity = init_param("humidity", 5, 20, 20, LINEAR, "%");
  Rn = init_param("Rn", 17, 0, 0.20, LOGSCALE, "");
#endif

  arg_index = handle_args(argc, argv);

  if(arg_index != argc-1) {
    printf("%s: wrong number of arguments\n", argv[0]);
    printusage();
    exit(EXIT_FAILURE);
  }
   
  /* Assign initial values to leaf parameters */
  /* M, b, alpha, littleGamma, Vm0 */

#define M_CHANGE 1
#define G_CHANGE 1

#if M_CHANGE
#warning "### TEMPORARY CHANGES TO M ###"
  c3lf.M = 20.0;
#else
  c3lf.M = 8.0;
#endif

  c3lf.b = 1.0e4;
  c3lf.alpha = 0.08;
#if G_CHANGE
#warning "### TEMPORARY CHANGES TO GAMMA ###"
  c3lf.littleGamma = 0.01;
#else
  c3lf.littleGamma = 0.02;
#endif
  c3lf.Vm0 = 12.5; /* par version from the other mech */

#if M_CHANGE  
  c4lf.M = 10.0;
#else
  c4lf.M = 4.0;
#endif

  c4lf.b = 0.01 * 1.0e6;
  c4lf.alpha = 0.06;
#if G_CHANGE
  c4lf.littleGamma = 0.02;
#else
  c4lf.littleGamma = 0.04;
#endif

  c4lf.Vm0 = 12.5;

  /* atmospheric CO2 concentration from commandline or default */
  theEnv.ca = co2_ppm * 1.0e-6;
  
  /* set the pointers */
  cp.lv = &cl;
  cp.ep = &theEnv;
  
  for(Cx = C_MIN; Cx <= C_MAX; Cx++) { /* C3 and C4 */
    fprintf(stderr, "%s\n", Cx ? "C4" : "C3");
    sprintf(binary_outfile_name, "mechlut.%s.c%d.bin", argv[arg_index], Cx+3);
    printf("binary output file name = %s\n", binary_outfile_name);

    if(BINARY_OUTPUT)
      outfile_binary = open_or_die(binary_outfile_name, "wb");
    if(ASCII_OUTPUT)
      outfile_ascii =  open_or_die(ascii_outfile_name, "w");
    
    print_file_header(Cx);
    print_param_header(CO2);
    print_param_header(Vm0);
    print_param_header(T);
    print_param_header(humidity);
    print_param_header(Rn);
    if(BINARY_OUTPUT)
      fprintf(outfile_binary, "DATA:\n");
    
    /* set leaf params to c3 or c4*/
    cp.lp = Cx ? &c4lf : &c3lf;

    /* progress bar */
    printf("progress:\n|");
#warning "progress bar semi-broken for large LUTs "
    for(i=0; i< CO2->N * Vm0->N; i++)
      printf(" ");
    printf("|\n");
    fflush(stdout);

    for(CO2_index = 0; CO2_index < CO2->N; CO2_index++){
      theEnv.ca = CO2->val[CO2_index] * 1.0e-6;
      /* loop through Vm0 levels */
      for(Vm0_index = 0; Vm0_index < Vm0->N; Vm0_index++) {
	  
	c3lf.Vm0 = Vm0->val[Vm0_index];
	c4lf.Vm0 = Vm0->val[Vm0_index];
	//printf("Vm0 bin = %d, Vm0 = %.2f\n", Vm0_index, Vm0); /* debug */
      
	for(T_index = 0; T_index < T->N; T_index++) {
	  theEnv.ta = T->val[T_index];

	  for(humidity_index=0; humidity_index<humidity->N; humidity_index++) {
	    /* convert humidity to mol/mol */
	    theEnv.ea = convert_humidity(humidity->val[humidity_index],
					 theEnv.ta);

	    for(Rn_index = 0; Rn_index < Rn->N; Rn_index++){

	      theEnv.L = Rn->val[Rn_index] * convRn2PAR * convW2mkEins;

	      theEnv.Rn = Rn->val[Rn_index];

	      /* solve */
	      status =  SolveFarquar(&soln, &cp, Cx);
	      
	      
	      switch (status) {
	      case SUCCESS:
		/* we're fine */
		break;
	      case FALLBACK:
		/* solver fell back to closed case. report it? */
		if(REPORT_FALLBACK) {
		  printf("solver fell back to closed case "
			 " (%e PAR %e RNET %e Ea %e Tf)"
			 " Falling back to the closed case.\n",
			 theEnv.L, theEnv.Rn, theEnv.ea, theEnv.ta);
		}
		break;
	      default:
		/* solver failed unrecoverably. return error conditon. */
		printf("Solver failed with status code %d:\n"
		       "%s CO2 %.2f Vm0 %.2f T %.2f hum %.2f par %.2f\n",
		       status, (Cx?"C4":"C3"), CO2->val[CO2_index],
		       Vm0->val[Vm0_index], T->val[T_index],
		       humidity->val[humidity_index],
		       Rn->val[Rn_index] );
		print_envPars(stdout, cp.ep);
		return 1;
	      }

#if 0	      
	      /* debug */
	      fprintf(outfile, "%d\t%f\t %f\t %f\t %f\t %f\t",
		      fallback, CO2->val[CO2_index], Vm0->val[Vm0_index],
		      T->val[T_index], humidity->val[humidity_index],
		      Rn->val[Rn_index]);
#endif
	      
	      print_result(soln.Ao, soln.Pso, soln.Ac, soln.Psc);

	      //fprintf(outfile,"\t%d\t%f\n", fallback, cl.tl);	/* debug */
	      //fprintf(outfile,"\t%f\n", cl.tl);	/* debug */
	      
	      /* reset all values to zero  */
	      cl.A = 0.0;
	      cl.Ps = 0.0;
	      cl.tl = 0.0;
	      cl.cs = 0.0;
	      cl.Ci = 0.0;
	      
	    } /* end Rn loop */
	    
	  } /* end humidity loop */
	  
	} /* end temperature loop */
      printf("=");		/* progress bar */
      } /* end Vm0 loop */
      fflush(stdout);
    } /* end CO2 loop */
    printf("|\n");
    if(in_file) fclose(in_file); 
    //if(outfile) fclose(outfile);
  } /* C3/C4 */

  /* debug */
  //  printf("Fell back on the closed case %d times\n", fallback);

  return 0;
} /* end of main function */


/** 
 * Try to open a file with the given mode. Exit if open fails.
 */
FILE * open_or_die(char* filename, char* mode) {
  FILE * file = NULL;
  file = fopen(filename, mode);
  if (file == NULL) {
    printf("Cannot open %s\n", filename);
    exit (EXIT_FAILURE);
  }
  return file;
}

/**
 * Convert islscp coordinates to ED and trim file extension
 */
static char * convert_islscp_filename(char * name) {
  float lat, lon;
  char * result;
  result = malloc(256*sizeof(char));
  
  name = basename(name);
  if(sscanf(name, "lat%flon%f.pin", &lat, &lon) != 2) {
    fprintf(stderr, "couldn't scan input filename: %s\n", name);
    printusage();
    exit(EXIT_FAILURE);
  }
  /* change coords from NW corner to center */
  lat -= 0.5;
  lon += 0.5;
  /* construct the result */
  snprintf(result, 256, "lat%.1flon%.1f", lat, lon );
  return result;
}
 
/** 
 * Prints args to binary and/or text output file.
 * 
 * @param Ao 
 * @param Pso 
 * @param Ac 
 * @param Psc 
 */
void print_result(double Ao, double Pso, double Ac, double Psc) {
  float tmp[4] = {Ao, Ac, Pso, Psc};
  if(BINARY_OUTPUT) {
    fwrite(tmp, sizeof(float), 4, outfile_binary);
  }
  if(ASCII_OUTPUT) {
    fprintf(outfile_ascii, "%e\t%e\t%e\t%e\n", Ao, Ac, Pso, Psc);
  }
}

/**
 * Convert % humidity (range 0-100) to a mixing ratio in
 * mol/mol at a given centigrade temperature.
 */
double convert_humidity(double h, double t) {
  t += ZeroC;			// convert to kelvin 
  return(exp(-5415.0/t) * 2.5414e6 * h/100.0);
}


void print_envPars(FILE *outfile, struct envPars *e) {
  fprintf(outfile, "envPars:\n");
  fprintf(outfile, "    T: %f (deg C)\n", e->ta);
  fprintf(outfile, "  CO2: %f (mol/mol)\n", e->ca);
  fprintf(outfile, " hum.: %f (mol/mol)\n", e->ea);
  fprintf(outfile, " absorbed PAR: %f (uEinsteins m^-2 s^-1)\n", e->L);
  fprintf(outfile, " net rad.: %f (J m^-2 s^-1)\n", e->Rn);
}

/** 
 * Print a header to the output file. It is in ascii, regardless of
 * whether output is binary
 * 
 * @param Cx 0 = C3, 1 = C4
 */
void print_file_header(int Cx) {
  if(BINARY_OUTPUT) {
    fprintf(outfile_binary, "%s\n", Cx ? "C4":"C3");
    fputs("variable\tN\tmin\tmax\tstep\tlogscale?\tunits\n", outfile_binary);
  }
  if(ASCII_OUTPUT) {
    fprintf(outfile_ascii, "%s\n", Cx ? "C4":"C3");
    fputs("variable\tN\tmin\tmax\tstep\tlogscale?\tunits\n", outfile_ascii);
    fputs("Ao\tAc\tPso\tPsc\n", outfile_ascii);
  }
}

void print_param_header(struct param *l) {
  if(BINARY_OUTPUT) {
    fprintf(outfile_binary, "%s\t%d\t%f\t%f\t%f\t%d\t%s\n",
	    l->name, l->N, l->min, l->max, l->step, l->logscale, l->units);
  }
  if(ASCII_OUTPUT) {
    fprintf(outfile_ascii, "%s\t%d\t%f\t%f\t%f\t%d\t%s\n",
	    l->name, l->N, l->min, l->max, l->step, l->logscale, l->units);    
  }
}
