/**
 * Set of functions to solve the farquar equations for a set of
 * environmental parameters that can be called from R With minor mods.
 * Can be set to be called from other pieces of C code.
 *
 * Written by Marco Albani <malbani@fas.harvard.edu> on April 22 2003
 * Based on previous work of David Medvigy and on routines printed in
 * the Numerical Recipes book.
 *
 * Copyright (C) 2003 The President and Fellows of Harvard College
 */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <libiberty.h>
#include "farquar.h"

/* Private function prototypes */
static int MySolverOpen(double tol, struct pars * thepar,
			double (*fx)(double, struct pars *));
static int MySolverClosed(double tol, struct pars * thepar,
			  double (*fx)(double, struct pars *));

/* ACTUAL FUNCTIONS  */

/* wrapper function that solves for both the open and the closed case */
int SolveFarquar(struct solution * thesol, struct pars * cp, int C3orC4)
{
  /* locals */
  struct pars  cpar1;
  struct pars  cpar2;

  struct pars * cp1;
  struct pars * cp2;

  struct envPars  ep1;
  struct envPars  ep2;

  struct leafVars  lv1;
  struct leafVars  lv2;

  struct leafPars  lp1;
  struct leafPars  lp2;


  double theTol;
  int theRet,theRet1, theRet2;
  int Cx;	/* 0 = C3 1 = C4 */

  /* assign substructure pointers */

  cpar1.lv =  &lv1;
  cpar1.ep = &ep1;
  cpar1.lp = &lp1;

  cpar2.lv =  &lv2;
  cpar2.ep = &ep2;
  cpar2.lp = &lp2;

  cp1 = &cpar1;
  cp2 = &cpar2;

  theTol =TOL;		/* set the tolerance to the macro specification */
  Cx = C3orC4; 
  /* set all values to zero  */
  thesol->Ao = 0.0;
  thesol->Pso = 0.0;
  thesol->tlo = 0.0;
  thesol->cso = 0.0;
  thesol->Cio = 0.0;
  thesol->Vmo = 0.0;
  thesol->Ac = 0.0;
  thesol->Psc = 0.0;
  thesol->tlc = 0.0;
  thesol->csc = 0.0;
  thesol->Cic = 0.0;
  thesol->Vmc = 0.0;

  cp->lv->A = 0.0;
  cp->lv->Ps = 0.0;
  cp->lv->tl = 0.0;
  cp->lv->cs = 0.0;
  cp->lv->Ci = 0.0;
  cp->lv->Vm = 0.0;

  /* solve for the closed case first */

  /* solve for the closed case first */

  theRet = MySolverClosed(theTol, cp, Cx ? AcC4:AcC3);
  if(theRet != 0){
    return FAILURE_CLOSED;
  }
  /* store the closed case set the solutions */
  thesol->Ac = cp->lv->A;
  thesol->Psc = cp->lv->Ps;
  thesol->tlc = cp->lv->tl;
  thesol->csc = cp->lv->cs;
  thesol->Cic = cp->lv->Ci;
  thesol->Vmc = cp->lv->Vm;

  /* reset leaf variables to zero */
  cp->lv->A = 0.0;
  cp->lv->Ps = 0.0;
  cp->lv->tl = 0.0;
  cp->lv->cs = 0.0;
  cp->lv->Ci = 0.0;
  cp->lv->Vm = 0.0;


  if(cp->ep->L < LIGHT_TOLERANCE) {
    /* below light tolerance, we are done */
    thesol->Ao  =  thesol->Ac;
    thesol->Pso =  thesol->Psc;
    thesol->tlo =  thesol->tlc;
    thesol->cso =  thesol->csc ;
    thesol->Cio =  thesol->Cic;
    thesol->Vmo =  thesol->Vmc;

    /* reset leaf variables to zero */
    cp->lv->A = 0.0;
    cp->lv->Ps = 0.0;
    cp->lv->tl = 0.0;
    cp->lv->cs = 0.0;
    cp->lv->Ci = 0.0;
    cp->lv->Vm = 0.0;

    return FALLBACK;
  } else { 			/* need to solve for the open case */
    if(Cx){
      /* C4 case, needs to solve separately for the three cases
	 (light, rubisco and co_2) */

      /* copy cp in cp1 and cp2 */
      copyPars(cp, cp1);
      copyPars(cp, cp2);

      /* solve for all three cases */
      /* and replace with the close case if the solution return a fallback  */
      theRet =  MySolverOpen(theTol, cp, AoC4light);
      if(theRet == FALLBACK){
	AcC4light(thesol->tlc, cp); /* fallback to the closed case */
      }else{
	if(theRet != SUCCESS){ 
	  fprintf(stderr, "AoC4light failed to find a solution with code %d\n",
		  theRet);
	}	/* report the failure */
#if DEBU_ROOTS
	AoC4test_light(cp->lv->tl,cp);
#endif
      }

      
      theRet1 =  MySolverOpen(theTol, cp1, AoC4rubisco);
      if(theRet1 == FALLBACK){
	AcC4rubisco(thesol->tlc, cp1); /* fallback to the closed case */
      }else{
	if(theRet1 != SUCCESS){
	  fprintf(stderr,
		  "AoC4rubisco failed to find a solution with code %d\n",
		  theRet1);
	}	/* report the failure */
#if DEBU_ROOTS
	AoC4test_rubisco(cp1->lv->tl,cp1);
#endif
      }



      theRet2 =  MySolverOpen(theTol, cp2, AoC4co_2);

      if(theRet2 == FALLBACK){
	AcC4co_2(thesol->tlc, cp2); /* fallback to the closed case */
      }else{
	if(theRet2 != SUCCESS){ 
	  fprintf(stderr, "AoC4co_2 failed to find a solution with code %d\n",
		  theRet2);
	}	/* report the failure */
#if DEBU_ROOTS
	AoC4test_co_2(cp2->lv->tl,cp2);
#endif
      }


      if((theRet == SUCCESS)||(theRet   == FALLBACK)) {
	/* Open case was successful */
	/* set the solutions */
	thesol->Ao = cp->lv->A;
	thesol->Pso = cp->lv->Ps;
	thesol->tlo = cp->lv->tl;
	thesol->cso = cp->lv->cs;
	thesol->Cio = cp->lv->Ci;
	thesol->Vmo = cp->lv->Vm;

	if((theRet1 == SUCCESS)||(theRet1   == FALLBACK)){
	  if(thesol->Ao > cp1->lv->A){
	    /* valid rubisco case has a lower A, use that one */
	    thesol->Ao = cp1->lv->A;
	    thesol->Pso = cp1->lv->Ps;
	    thesol->tlo = cp1->lv->tl;
	    thesol->cso = cp1->lv->cs;
	    thesol->Cio = cp1->lv->Ci;
	    thesol->Vmo = cp1->lv->Vm;
	  }
	}

	if(((theRet2 == SUCCESS)||( theRet2  == FALLBACK))&&(thesol->Ao > cp2->lv->A)){
	  /* valid co_2 case has a lower A, use that one */
	  thesol->Ao = cp1->lv->A;
	  thesol->Pso = cp1->lv->Ps;
	  thesol->tlo = cp1->lv->tl;
	  thesol->cso = cp1->lv->cs;
	  thesol->Cio = cp1->lv->Ci;
	  thesol->Vmo = cp1->lv->Vm;
	}

	/* we are done, return */
	return SUCCESS;
      }else{
	/* Light case has failed, use the rubisco cases */
	if((theRet1 == SUCCESS)||(theRet1   == FALLBACK)){
	  thesol->Ao = cp1->lv->A;
	  thesol->Pso = cp1->lv->Ps;
	  thesol->tlo = cp1->lv->tl;
	  thesol->cso = cp1->lv->cs;
	  thesol->Cio = cp1->lv->Ci;
	  thesol->Vmo = cp1->lv->Vm;

	  if(((theRet2 == SUCCESS)||( theRet2 == FALLBACK))&&(thesol->Ao > cp2->lv->A)){
	    /* valid co_2 case has a lower A, use that one */
	    thesol->Ao = cp1->lv->A;
	    thesol->Pso = cp1->lv->Ps;
	    thesol->tlo = cp1->lv->tl;
	    thesol->cso = cp1->lv->cs;
	    thesol->Cio = cp1->lv->Ci;
	    thesol->Vmo = cp1->lv->Vm;
	  }

	  /* we are done, return */
	  return SUCCESS;


	}else{
	  /* rubisco has failed too, Report and use the co_2 case */
	  if((theRet2 == SUCCESS)||( theRet2  == FALLBACK)){
	    thesol->Ao = cp2->lv->A;
	    thesol->Pso = cp2->lv->Ps;
	    thesol->tlo = cp2->lv->tl;
	    thesol->cso = cp2->lv->cs;
	    thesol->Cio = cp2->lv->Ci;
	    thesol->Vmo = cp2->lv->Vm;

	    /* we are done, return */
	    return SUCCESS;
	  }else{
	    return FAILURE_OPEN; /* everything has failed, return failure */
	  }
	}
      }
      

    }else{
      /* C3 case, needs to solve separately for the two cases
	 (light, and rubisco) */

      /* copy cp in cp1 */
      copyPars(cp, cp1);


      /* solve for all two cases */
      /* and replace with the close case if the solution return a fallback  */
      theRet =  MySolverOpen(theTol, cp, AoC3light);
      if(theRet == FALLBACK){
	AcC3light(thesol->tlc, cp); /* fallback to the closed case */
      }else{
	if(theRet != SUCCESS){ 
	  fprintf(stderr, "AoC3light failed to find a solution with code %d\n",
		  theRet);
	}	/* report the failure */
#if DEBU_ROOTS
	AoC3test_light(cp->lv->tl,cp);
#endif
      }

      
      theRet1 =  MySolverOpen(theTol, cp1, AoC3rubisco);
      if(theRet1 == FALLBACK){
	AcC3rubisco(thesol->tlc, cp1); /* fallback to the closed case */
      }else{
	if(theRet1 != SUCCESS){
	  fprintf(stderr,
		  "AoC3rubisco failed to find a solution with code %d\n",
		  theRet1);
	}	/* report the failure */
#if DEBU_ROOTS
	AoC3test_rubisco(cp1->lv->tl,cp1);
#endif
      }


      if((theRet == SUCCESS)||(theRet   == FALLBACK)) {
	/* Open case was successful */
	/* set the solutions */
	thesol->Ao = cp->lv->A;
	thesol->Pso = cp->lv->Ps;
	thesol->tlo = cp->lv->tl;
	thesol->cso = cp->lv->cs;
	thesol->Cio = cp->lv->Ci;
	thesol->Vmo = cp->lv->Vm;

	if((theRet1 == SUCCESS)||(theRet1   == FALLBACK)){
	  if(thesol->Ao > cp1->lv->A){
	    /* valid rubisco  case has a lower A, use that one */
	    thesol->Ao = cp1->lv->A;
	    thesol->Pso = cp1->lv->Ps;
	    thesol->tlo = cp1->lv->tl;
	    thesol->cso = cp1->lv->cs;
	    thesol->Cio = cp1->lv->Ci;
	    thesol->Vmo = cp1->lv->Vm;
	  }
	}
	/* we are done, return */
	return SUCCESS;
      }else{ 			/* Light case has failed, use the rubisco cases */
	if((theRet1 == SUCCESS)||(theRet1   == FALLBACK)){
	  thesol->Ao = cp1->lv->A;
	  thesol->Pso = cp1->lv->Ps;
	  thesol->tlo = cp1->lv->tl;
	  thesol->cso = cp1->lv->cs;
	  thesol->Cio = cp1->lv->Ci;
	  thesol->Vmo = cp1->lv->Vm;

	  /* we are done, return */
	  return SUCCESS;

	}else{
	  return FAILURE_OPEN; /* everything has failed, return failure */
	}
      }
      
    } /* end if  C3 open case */
  } /* end if light limited case  */
} /* end function */

/* the zbrak function */
int MyZbrak(double (*fx)(double, struct pars *), double x1, double x2, int n, double xb1[], double xb2[], int *nb, struct pars * thePars)
{
  int nbb,i;
  double x,fp,fc,dx;


  nbb=0;
  dx=(x2-x1)/n;
  x=x1;
  fp=(*fx)(x, thePars);
  for(i=0; i < n; i++) {
    /* change from numerical recipes to get C-style arrays (0 to n-1) */
    x += dx;
    fc=(*fx)(x, thePars);
    if(fc*fp <= 0.0) {
      xb1[nbb]=x-dx;
      xb2[nbb]=x;
#if DEBU_ROOTS
      printf("MyZbrak found a solution between %f and %f.\n", xb1[nbb], xb2[nbb]);
#endif

      nbb++;
      if(*nb == nbb) return 0; 	/* found the max number of solutions */
    }
    fp=fc;
  }
  *nb = nbb;
  return 0;
}


/* the MyZbrent function */
/* Using Brent's method find the root of a function func known to lie
   between x1 and x2. The root, returned as zbrent, will be redifined
   until its accuracy is tol */

double MyZbrent(double (*func)(double, struct pars *), double x1, double x2, double tol, struct pars * thePars){
  int iter;
  double a=x1, b=x2, c=x2, d=0.0, e=0.0, min1, min2;
  double fa=(*func)(a, thePars), fb=(*func)(b, thePars), fc, p, q, r,s,tol1,xm;
#if DEBU_ROOTS
  printf("MyZbrent Starting between %f and %f\n", x1,x2);
#endif
  if ((fa > 0.0 && fb > 0.0) || (fa < 0.0 && fb < 0.0)) {
#if DEBU_ROOTS
    printf("%s\n", "MyZbrent Error: Root Must be bracketed!"); 
#endif
  }
  fc = fb;

  for(iter=1; iter<=ITMAX; iter++) {
    if ((fb > 0.0 && fc > 0.0) || (fb < 0.0 && fc < 0.0)){
      c = a;	   /* rename a,b and c and adjust bounding interval */
      fc = fa;
      d =  b-a;
      e = d;
    }
    if(myFabs(fc) < myFabs(fb)) {
      a=b;
      b=c;
      c=a;
      fa=fb;
      fb=fc;
      fc=fa;
    }
    tol1 = 2.0 * EPS * myFabs(b) + 0.5 * tol; /* convergence check */
    xm = 0.5*(c-b);
    if(myFabs(xm) <= tol1 || fb == 0.0) return(b);
    if((myFabs(e) >= tol1) && (myFabs(fa) > myFabs(fb))){
      s=fb/fa;		 /* attempt inverse quadratic interpolation */
      if (a == c) {
	p=2.0*xm*s;
	q=1.0-s;
      }else{
	q=fa/fc;
	r=fb/fc;
	p=s*(2.0*xm*q*(q-r)-(b-a)*(r-1.0));
	q=(q-1.0)*(r-1.0)*(s-1.0);
      }
      if (p > 0.0) q = -q;
      p=myFabs(p);
      min1=3.0*xm*q-myFabs(tol1*q);
      min2=myFabs(e*q);
      if(2.0*p < ((min1 < min2) ? min1 : min2)){ /* accept interpolation */
	e=d;
	d=p/q;
      }else{
	d=xm;		    /* interpolation failed, used bisection */
	e=d;
      }
    }else{
      /* bounds decreasing too slowly, use bisection */
      d=xm;
      e=d;
    }
    a=b;			/* move last best guess to a */
    fa=fb;

    if(myFabs(d) > tol1)	/* evaluate new trial root */
      b +=d;
    else
      b+= SIGN(tol1,xm);

    fb=(*func)(b, thePars);
  }
#if DEBU_ROOTS
  printf("%s \n","MyZbrent Error: Maximum number of iterations exceeded in zbrent!");
#endif

  return 0.0;			/* Never get here, return zero Kelvin */
}


/* TfClosed is a function that solves for the temperature only in the
   closed case */

double TfClosed(double Tf, struct pars * p){
  double Ttest;


#if DEBU_FUNCTIONS

  printf("TfClosed starting\n");
  printf("\n");

#endif

  /* solve immediately for Tf */
  Ttest = p->ep->Rn * (1-exp(-0.5)) -  Theta*(Tf - p->ep->ta) - Lambda * p->lp->b *
    ( (2.5414e6) * exp( -5415.0 / (Tf + ZeroC)) - p->ep->ea);

  p->lv->tl = Tf;

  return Ttest; /* returns the departure from 0 in the solution */
}





double myFabs(double arg){

  if(arg < 0.0) arg = -arg;
  return(arg);
}


static int MySolverOpen(double tol, struct pars *p, double (*fx)(double, struct pars *))
{
  double x1;
  double  x2;
  int theRet = 0;
  int n = 10;		       /* number of brackets with solutions */
  int nbr = 10;			/* maximum number of brackets */
  int i, fallback = 0;
  double theTf, theZero;
  double *xb1; /* arrays with bracketed solutions, declared as pointers */
  double *xb2;
  int Success = 0;		/* test for success */

  /* limits for testing solutions */
  double maxCi = 1.0;
  double minCi = 1.0e-8;
  double minCs = -1.0e7;  /* fail for absurdly low Cs - other Cs < b
			     cause fallback to te closed case */
  double maxCs = 1.0e8;

#if DEBU_SOLVER

  printf("MySolverOpen starting\n");
  printf("\n");

#endif


  /* first set the initial bracketing interval based on the air temperature */
  x1 = 0.0 - 12.0;		/* -12 degrees C below */
  x2 = 60.0;			/* 60 degrees C  */

  fallback = 0;

  while(Success == 0) {
    /* allocate nbr elements to the arrays that hold the bracket extremes */
    xb1 = (double *)xmalloc(nbr * sizeof(double));
    xb2 = (double *)xmalloc(nbr * sizeof(double));
    /* now find the brackets */
 
    theRet = MyZbrak((*fx), x1, x2,  n, xb1, xb2, &nbr, p);

    if(nbr < 1) {	/* no solutions where found in the brackets */
      theRet = 1;
#if DEBU_SOLVER
      printf("MyZbrak() failed to find any solution brackets using %i intervals"
	     "\nIncreasing the number of intervals to %i\n", n, n*10);
#endif
      n = n * 10;     /* increase the number of brackets by tenfold */
      nbr = n;
      if(nbr > MAX_NBR){
	printf("MyZbrak() failed to find any solution brackets using %i intervals\nBailing out.... \n\n", nbr);
	free(xb1);
	free(xb2);
	return BRACKET_FAILURE;
      }
    } else {
      /* there was at least a bracket with at least one solution */
      /* loop through the brackets */
      for (i = 0; i < nbr; i++){
	theTf = MyZbrent((*fx),xb1[i],xb2[i], tol, p);
#if DEBU_SOLVER
	printf("Zbrenr returned %f and leaf temperature in the structure is %f.\n", theTf, p->lv->tl);
#endif
	/* update the structure using the Tf returned (Zbrent doesn't
	   leave the structure clean!!!)  */
	theZero = (*fx)(theTf,p);
#if DEBU_SOLVER
	printf("Now theTf is  %f and leaf temperature in the structure is %f.\n", theTf, p->lv->tl);
#endif
	if(((p->lv->Ci < maxCi) &&(p->lv->Ci > minCi)) && 
	   ((p->lv->cs < maxCs) &&(p->lv->cs > minCs))){
	  /* success!! we found a viable solution */
	  if(p->lv->cs >= p->lp->b){
	    Success = 1;
#if DEBU_SOLVER
	    printf("success! Found a solution for Tf of %f the root is  %e.\n", theTf, theZero);
#endif
	    break;		/* exit the for loop */
	  } else {
	    /* we found a solution but it is wrong, got to fall back
	       to the closed case */

#if DEBU_SOLVER
	    printf("Found a solution for Tf of %f, but the cs is %e (less than b0)\n", theTf,p->lv->cs );
	    printf("Setting fallback to 1\n");
#endif
	    fallback = 1;
	  }
#if DEBU_SOLVER
	} else {

	  printf("Cs is  %f and Ci %f. - Non-valid solution\n", p->lv->cs, p->lv->Ci);

#endif
	}/* end if */
      } /* end for loop */
      if(Success == 0) {
	theRet = 2;
#if DEBU_SOLVER
	printf("MyZbrent() found non viable solution using within %i intervals\nIncreasing the number of intervals to %i\n", n, n*10);
#endif
	n = n * 10; 		/* increase the number of brackets by tenfold */
	nbr = n;
	if(nbr > MAX_NBR
	   /* don't search as much if we already have a fallback */
	   || (fallback == 1 && nbr > MAX_NBR/10))
	  {
#if DEBU_SOLVER
	    printf("MyZbrent() failed to find any viable solutions in the interval using %i brackets \nBailing out...\n", n);
#endif
	    free(xb1);
	    free(xb2);
	    if(fallback == 1){
	      return FALLBACK;
	    } else {
	      return CONVERGE_FAILURE;
	    }
	  } /* endif */
      }
    } /* end of the else */
    
    free(xb1);
    free(xb2);			/* free the memory allocated */
  } /* end of the while loop */
  /* We have a solution, we can get out */
  return 0;
}

/* function sign to return the sign of a double */
double sign(double x)
{
  if(x >= 0) {
    return 1.0;
  }else{
    return -1.0;
  }
}

static int MySolverClosed(double tol, struct pars *p, double (*fx)(double, struct pars *))
{
  double x1;
  double  x2;
  double theTf, theZero;


#if DEBU_SOLVER
  printf("MySolverClosed starting\n");
#endif


  /* set the initial bracketing interval based on the air temperature */
  x1 = 0.0 -12.0;		/* -12 degrees C below air temp */
  x2 = 60.0;			/* 60 degrees C above air temp */

  /* just go ahead and call Brent's method, the function is well behaved  */
  theTf = MyZbrent(TfClosed,x1,x2, tol, p);

#if DEBU_SOLVER
  printf("Zbrent returned %f and leaf temperature in the structure is %f.\n", theTf, p->lv->tl);
#endif
  /* update the structure using the Tf returned (Zbrent doesn't leave
     the structure clean!!!)  */
  theZero = (*fx)(theTf,p);
#if DEBU_SOLVER
  printf("Now theTf is  %f and leaf temperature in the structure is %f.\n", theTf, p->lv->tl);
  printf("success! Found a solution for Tf of %f the root is  %e.\n", theTf, theZero);
#endif
  /* We have a solution, we can get out */
  return SUCCESS;
}

/* ********************************************* */
/*                  C3, Closed Cases             */
/* ********************************************* */

/* C3 Plant, A closed */
double AcC3(double Tf, struct pars * p){

#if CLOSED_CASE_ASSIMILATION
  double arubisco, alight;
  double Gamma;			/* CO2 compensation point [mol/mol] */
  double K1, K2;		/* Parameters for dark reaction */
  double a, b1, b2, c1, c2, q1 , q2; /* Parameters of the quadratic eq for Ci) */
  double Ci1, Ci2;	       /* Ci light case and Ci rubisco case */
#endif
  /* note that there are calls to the physical constants Theta, Lamda
     and D0 that are defined globally */

#if DEBU_FUNCTIONS

  printf("AcC3 starting\n");
  printf("\n");

#endif

  /* update the leaf variables */

  p->lv->cs = p->lp->b;

  p->lv->Ps = (p->ep->Rn * (1-exp(-kappa)) - Theta * (Tf - p->ep->ta))/Lambda;

  p->lv->Vm = p->lp->Vm0 * exp(3000.0 *(1.0/288.2 - 1.0/(Tf + ZeroC)))/
    (1.0 + exp(0.4 * (5.0 -Tf)))/
    (1.0 + exp(0.4 * (Tf - 45.0)));
#if CLOSED_CASE_ASSIMILATION	
  K1 = (1.5e-4) * exp(6000.0 *(1.0/288.2 - 1.0/(Tf + ZeroC)));

  K2 = 0.836 * exp(-1400.0 *(1.0/288.2 - 1.0/(Tf + ZeroC)));


  Gamma = (2.12e-5) * exp(5000.0*(1.0/288.2 - 1.0/(Tf+ZeroC)));

  /* solve for Ci */

  
  a = -p->lp->b/1.6;
  /* by definition a is always negative */

  /* b terms can be either postive or negative */
  b1 = p->lp->b * p->ep->ca/1.6 + p->lv->Vm * p->lp->littleGamma - p->lp->alpha * p->ep->L * (1-exp(-0.5))
    -  p->lp->b *  Gamma/1.6;

  b2 =  p->lp->b * p->ep->ca/1.6 + p->lv->Vm * p->lp->littleGamma - p->lp->b * K1 * (1+K2)/1.6 - p->lv->Vm;


  /* c terms are always positive (sum and product of positive terms) */
  c1 = Gamma * (p->lp->b *  p->ep->ca/1.6 +  p->lv->Vm * p->lp->littleGamma +  p->lp->alpha * p->ep->L
		* (1-exp(-0.5)));

  c2 = K1*(1+K2)*(p->lp->b *  p->ep->ca/1.6 +  p->lv->Vm * p->lp->littleGamma)+ p->lv->Vm*Gamma;

  /* q terms have sign always opposite to the corresponding b term */
  q1 = -0.5 * ( b1 + sign(b1) * sqrt(pow(b1, 2) - 4 * a * c1));
  q2 = -0.5 * ( b2 + sign(b2) * sqrt(pow(b2, 2) - 4 * a * c2));

  /* check the solutions and select one  */
  /* note that when one of the solution is positive the other must be
     negative. this because a and the c terms must have opposite signs
     and q has sign opposite to b (the square root term is always
     positive). the approach below also handles tha case where q is 0
     (shouldn't happen, but you never know) avoiding divison by 0 */

  /* roots for the light case */
  if(q1 <= 0){
    Ci1 = q1/a;
  }else{
    Ci1 = c1/q1;
  }
  
  /* roots for the rubisco case */

  if(q2 <= 0){
    Ci2 = q2/a;			/* this root is positive */
  }else{
    Ci2 = c2/q2;		/* this other root is positive */
  }

#if DEBU_FUNCTIONS 
  printf("AcC3: Ci roots are %f (light case) and %f (rubisco case)\n",Ci1,Ci2); 
  printf("AcC3: Vm is %f - Gamma is: %f ",  p->lv->Vm,
	 Gamma); 
#endif

  arubisco = p->lv->Vm * ( Ci2 - Gamma)/( Ci2 + K1*(1.0 + K2));
	
	
  alight = p->lp->alpha * p->ep->L  * (1-exp(-0.5)) * (( Ci1 - Gamma)/( Ci1 + Gamma));
  //	alight = p->lp->alpha * p->ep->L  * (( p->lv->Ci - Gamma)/( p->lv->Ci + Gamma));	

#if DEBU_FUNCTIONS 
  printf("AcC3: A gross are %f (light case) and %f (rubisco case)\n", alight, arubisco); 
#endif

  if(arubisco <= alight){ /*minimum*/
    p->lv->A =  arubisco;
    p->lv->Ci = Ci2;
  }else{
    p->lv->A = alight;
    p->lv->Ci = Ci1;
  }
		
  if(p->lv->A < 0) p->lv->A = 0.0;
#if DEBU_FUNCTIONS 
  printf("AcC3: A gross is set to %f (min of rubiscoand light case)\n",p->lv->A); 
#endif

#else

  p->lv->A = 0.0;	     /* gross assimilation is identically 0 */

#endif


  p->lv->A -= p->lv->Vm * p->lp->littleGamma;
  p->lv->Ci = p->ep->ca -  (1.6 * p->lv->A)/p->lv->cs;

  p->lv->tl = Tf;

  return p->lv->A; /* returns the A */

}



/* C3 Plant, A closed, light case only */
double AcC3light(double Tf, struct pars * p){

#if CLOSED_CASE_ASSIMILATION
  double arubisco, alight;
  double Gamma;			/* CO2 compensation point [mol/mol] */
  double K1, K2;		/* Parameters for dark reaction */
  double a, b1, b2, c1, c2, q1 , q2; 	/* Parameters of the quadratic eq for Ci) */
  double Ci1, Ci2; 		/* Ci light case and Ci rubisco case */
#endif
  /* note that there are calls to the physical constants Theta, Lamda
     and D0 that are defined globally */

#if DEBU_FUNCTIONS

  printf("AcC3light starting\n");
  printf("\n");

#endif

  /* update the leaf variables */

  p->lv->cs = p->lp->b;

  p->lv->Ps = (p->ep->Rn * (1-exp(-kappa)) - Theta * (Tf - p->ep->ta))/Lambda;

  p->lv->Vm = p->lp->Vm0 * exp(3000.0 *(1.0/288.2 - 1.0/(Tf + ZeroC)))/
    (1.0 + exp(0.4 * (5.0 -Tf)))/
    (1.0 + exp(0.4 * (Tf - 45.0)));
#if CLOSED_CASE_ASSIMILATION	
  K1 = (1.5e-4) * exp(6000.0 *(1.0/288.2 - 1.0/(Tf + ZeroC)));

  K2 = 0.836 * exp(-1400.0 *(1.0/288.2 - 1.0/(Tf + ZeroC)));


  Gamma = (2.12e-5) * exp(5000.0*(1.0/288.2 - 1.0/(Tf+ZeroC)));

  /* solve for Ci */
  
  a = -p->lp->b/1.6;
  /* by definition a is always negative */

  /* b terms can be either postive or negative */
  b1 = p->lp->b * p->ep->ca/1.6 + p->lv->Vm * p->lp->littleGamma - p->lp->alpha * p->ep->L * (1-exp(-0.5))
    -  p->lp->b *  Gamma/1.6;

  b2 =  p->lp->b * p->ep->ca/1.6 + p->lv->Vm * p->lp->littleGamma - p->lp->b * K1 * (1+K2)/1.6 - p->lv->Vm;

  /* c terms are always positive (sum and product of positive terms) */
  c1 = Gamma * (p->lp->b *  p->ep->ca/1.6 +  p->lv->Vm * p->lp->littleGamma +  p->lp->alpha * p->ep->L
		* (1-exp(-0.5)));

  c2 = K1*(1+K2)*(p->lp->b *  p->ep->ca/1.6 +  p->lv->Vm * p->lp->littleGamma)+ p->lv->Vm*Gamma;

  /* q terms have sign always opposite to the corresponding b term */
  q1 = -0.5 * ( b1 + sign(b1) * sqrt(pow(b1, 2) - 4 * a * c1));
  q2 = -0.5 * ( b2 + sign(b2) * sqrt(pow(b2, 2) - 4 * a * c2));

  /* check the solutions and select one  */
  /* note that when one of the solution is positive the other must be
   negative this because a and the c terms must have opposite signs
   and q has sign opposite to b (the square root term is always
   positive) the approach below also handles tha case where q is 0
   (shouldn't happen, but you never know) avoiding divison by 0 */

  /* roots for the light case */
  if(q1 <= 0){
    Ci1 = q1/a;
  }else{
    Ci1 = c1/q1;
  }
  

#if DEBU_FUNCTIONS 
  printf("AcC3light: Ci root is %f \n",Ci1); 
  printf("AcC3light: Vm is %f - Gamma is: %f ",  p->lv->Vm,
	 Gamma); 
#endif


	
	
  alight = p->lp->alpha * p->ep->L  * (1-exp(-0.5)) * (( Ci1 - Gamma)/( Ci1 + Gamma));
  //	alight = p->lp->alpha * p->ep->L  * (( p->lv->Ci - Gamma)/( p->lv->Ci + Gamma));	

#if DEBU_FUNCTIONS 
  printf("AcC3light: A gross is %f (light case)\n", alight); 
#endif

  p->lv->A = alight;
  p->lv->Ci = Ci1;

  if(p->lv->A < 0) p->lv->A = 0.0;
#if DEBU_FUNCTIONS 
  printf("AcC3light: A gross is set to %f\n",p->lv->A); 
#endif

#else

  p->lv->A = 0.0; 		/* gross assimilation is identically 0 */

#endif

  p->lv->A -= p->lv->Vm * p->lp->littleGamma;
  p->lv->Ci = p->ep->ca -  (1.6 * p->lv->A)/p->lv->cs;

  p->lv->tl = Tf;

  return p->lv->A; /* returns the A */

}

/* C3 Plant, A closed */
double AcC3rubisco(double Tf, struct pars * p){

#if CLOSED_CASE_ASSIMILATION
  double arubisco, alight;
  double Gamma;			/* CO2 compensation point [mol/mol] */
  double K1, K2;		/* Parameters for dark reaction */
  double a, b1, b2, c1, c2, q1 , q2; 	/* Parameters of the quadratic eq for Ci) */
  double Ci1, Ci2; 		/* Ci light case and Ci rubisco case */
#endif
  /* note that there are calls to the physical constants Theta, Lamda
     and D0 that are defined globally */

#if DEBU_FUNCTIONS

  printf("AcC3rubisco starting\n");
  printf("\n");

#endif

  /* update the leaf variables */

  p->lv->cs = p->lp->b;

  p->lv->Ps = (p->ep->Rn * (1-exp(-kappa)) - Theta * (Tf - p->ep->ta))/Lambda;

  p->lv->Vm = p->lp->Vm0 * exp(3000.0 *(1.0/288.2 - 1.0/(Tf + ZeroC)))/
    (1.0 + exp(0.4 * (5.0 -Tf)))/
    (1.0 + exp(0.4 * (Tf - 45.0)));
#if CLOSED_CASE_ASSIMILATION	
  K1 = (1.5e-4) * exp(6000.0 *(1.0/288.2 - 1.0/(Tf + ZeroC)));

  K2 = 0.836 * exp(-1400.0 *(1.0/288.2 - 1.0/(Tf + ZeroC)));

  Gamma = (2.12e-5) * exp(5000.0*(1.0/288.2 - 1.0/(Tf+ZeroC)));

  /* solve for Ci */
  
  a = -p->lp->b/1.6;
  /* by definition a is always negative */

  /* b terms can be either postive or negative */
  b1 = p->lp->b * p->ep->ca/1.6 + p->lv->Vm * p->lp->littleGamma - p->lp->alpha * p->ep->L * (1-exp(-0.5))
    -  p->lp->b *  Gamma/1.6;

  b2 =  p->lp->b * p->ep->ca/1.6 + p->lv->Vm * p->lp->littleGamma - p->lp->b * K1 * (1+K2)/1.6 - p->lv->Vm;


  /* c terms are always positive (sum and product of positive terms) */
  c1 = Gamma * (p->lp->b *  p->ep->ca/1.6 +  p->lv->Vm * p->lp->littleGamma +  p->lp->alpha * p->ep->L
		* (1-exp(-0.5)));

  c2 = K1*(1+K2)*(p->lp->b *  p->ep->ca/1.6 +  p->lv->Vm * p->lp->littleGamma)+ p->lv->Vm*Gamma;

  /* q terms have sign always opposite to the corresponding b term */
  q1 = -0.5 * ( b1 + sign(b1) * sqrt(pow(b1, 2) - 4 * a * c1));
  q2 = -0.5 * ( b2 + sign(b2) * sqrt(pow(b2, 2) - 4 * a * c2));

  /* check the solutions and select one  */
  /* note that when one of the solution is positive the other must be
   negative. this because a and the c terms must have opposite signs
   and q has sign opposite to b (the square root term is always
   positive).  the approach below also handles tha case where q is 0
   (shouldn't happen, but you never know) avoiding divison by 0 */

  /* roots for the rubisco case */

  if(q2 <= 0){
    Ci2 = q2/a; 			/* this root is positive */
  }else{
    Ci2 = c2/q2;		/* this other root is positive */
  }

#if DEBU_FUNCTIONS 
  printf("AcC3rubisco: Ci root is  %f",Ci2); 
  printf("AcC3rubisco: Vm is %f - Gamma is: %f ",  p->lv->Vm,
	 Gamma); 
#endif

  arubisco = p->lv->Vm * ( Ci2 - Gamma)/( Ci2 + K1*(1.0 + K2));

#if DEBU_FUNCTIONS 
  printf("AcC3rubisco: A gross is %f\n",  arubisco); 
#endif

  p->lv->A =  arubisco;
  p->lv->Ci = Ci2;
		
  if(p->lv->A < 0) p->lv->A = 0.0;
#if DEBU_FUNCTIONS 
  printf("AcC3rubisco: A gross is set to %f \n",p->lv->A); 
#endif

#else

  p->lv->A = 0.0; 		/* gross assimilation is identically 0 */

#endif

  p->lv->A -= p->lv->Vm * p->lp->littleGamma;
  p->lv->Ci = p->ep->ca -  (1.6 * p->lv->A)/p->lv->cs;

  p->lv->tl = Tf;

  return p->lv->A; /* returns the A */

}

/* ********************************************* */
/*                  C4, Closed Cases             */
/* ********************************************* */

/* C4 Plant, A closed (single solution for the solver */

double AcC4(double Tf, struct pars * p){

  double Ttest;
#if CLOSED_CASE_ASSIMILATION
  double  arubisco, alight;
  double Gamma;			/* CO2 compensation point [mol/mol] */
  double Cil, Cir1, Cir2; 		/* Ci light case and Ci rubisco case (2) */
#endif
  /* note that there are calls to the physical constants Theta, Lamda
     and D0 that are defined globally */


#if DEBU_FUNCTIONS

  printf("AcC4 starting\n");
  printf("\n");

#endif

  /* soleve immediately for Tf */
  Ttest = p->ep->Rn * (1-exp(-kappa)) -  Theta*(Tf - p->ep->ta) - Lambda * p->lp->b *
    ( (2.5414e6) * exp( -5415.0 / (Tf + ZeroC)) - p->ep->ea);

  /* update the leaf variables */
	

  p->lv->cs = p->lp->b;

  p->lv->Ps = (p->ep->Rn * (1-exp(-kappa)) - Theta * (Tf - p->ep->ta))/Lambda;

  p->lv->Vm = p->lp->Vm0 * exp(3000.0 *(1.0/288.2 - 1.0/(Tf + ZeroC)))/
    (1.0 + exp(0.4 * (5.0 -Tf)))/
    (1.0 + exp(0.4 * (Tf - 100.0)));
#if CLOSED_CASE_ASSIMILATION

  /* solve for Ci */
  alight = p->lp->alpha * p->ep->L * (1-exp(-0.5));

  arubisco = p->lv->Vm;

  Cir2 = ( p->lv->Vm * p->lp->littleGamma + p->lp->b * p->ep->ca/1.6)/(p->lv->Vm * 18000 + p->lp->b/1.6);

	
  if( Cir2*18000 < 1.0 ){
    p->lv->Ci =  Cir2;
    arubisco = arubisco * 18000 * p->lv->Ci; /* arubisco min of Vm and
						18000*Vm*Ci */
  } else {

    Cir1 = 1.6/p->lp->b * (p->lv->Vm * p->lp->littleGamma + p->lp->b * p->ep->ca/1.6-
			   p->lv->Vm);
    p->lv->Ci =  Cir1;
	
  }
	
  if(arubisco <= alight){
    p->lv->A = arubisco;
  }else{
    p->lv->A = alight;
    Cil = 1.6/p->lp->b * ( p->lv->Vm * p->lp->littleGamma + p->lp->b * p->ep->ca/1.6 - alight);
    p->lv->Ci =  Cil;
  }

  if(p->lv->A < 0) p->lv->A = 0.0;
#else

  p->lv->A = 0.0; 		/* gross assimilation is identically 0 */

#endif

  p->lv->A -= p->lv->Vm * p->lp->littleGamma;
  p->lv->Ci = p->ep->ca -  (1.6 * p->lv->A)/p->lv->cs;

  p->lv->tl = Tf;

  return p->lv->A; /* returns the net absorption */

}

/* C4 Plant, A closed, light case */
double AcC4light(double Tf, struct pars * p){

  double Ttest;
#if CLOSED_CASE_ASSIMILATION
  double alight;
  double Gamma;			/* CO2 compensation point [mol/mol] */
  double Cil, Cir1, Cir2;  /* Ci light case and Ci rubisco case (2) */
#endif
  /* note that there are calls to the physical constants Theta, Lamda
     and D0 that are defined globally */

#if DEBU_FUNCTIONS

  printf("AcC4 starting\n");
  printf("\n");

#endif

  /* soleve immediately for Tf */
  Ttest = p->ep->Rn * (1-exp(-kappa)) -  Theta*(Tf - p->ep->ta) - Lambda * p->lp->b *
    ( (2.5414e6) * exp( -5415.0 / (Tf + ZeroC)) - p->ep->ea);

  /* update the leaf variables */

  p->lv->cs = p->lp->b;

  p->lv->Ps = (p->ep->Rn * (1-exp(-kappa)) - Theta * (Tf - p->ep->ta))/Lambda;

  p->lv->Vm = p->lp->Vm0 * exp(3000.0 *(1.0/288.2 - 1.0/(Tf + ZeroC)))/
    ((1.0 + exp(0.4 * (5.0 -Tf)))*
     (1.0 + exp(0.4 * (Tf - 100.0))));

#if CLOSED_CASE_ASSIMILATION

  /* solve for Ci */
  alight = p->lp->alpha * p->ep->L * (1-exp(-0.5));

  p->lv->A = alight;
  Cil = 1.6/p->lp->b * ( p->lv->Vm * p->lp->littleGamma + p->lp->b * p->ep->ca/1.6 - alight);
  p->lv->Ci =  Cil;

  if(p->lv->A < 0) p->lv->A = 0.0;
#else

  p->lv->A = 0.0; 		/* gross assimilation is identically 0 */

#endif

  p->lv->A -= p->lv->Vm * p->lp->littleGamma;
  p->lv->Ci = p->ep->ca -  (1.6 * p->lv->A)/p->lv->cs;


  p->lv->tl = Tf;

  return p->lv->A; /* returns the net absorption */

}

/* C4 Plant, A closed, unlimited rubisco case */
double AcC4rubisco(double Tf, struct pars * p){

  double Ttest;
#if CLOSED_CASE_ASSIMILATION
  double  arubisco;
  double Gamma;			/* CO2 compensation point [mol/mol] */
  double Cir1; 		/* Ci light case and Ci rubisco case (2) */
#endif
  /* note that there are calls to the physical constants Theta, Lamda
     and D0 that are defined globally */

#if DEBU_FUNCTIONS

  printf("AcC4 starting\n");
  printf("\n");

#endif

  /* solve immediately for Tf */
  Ttest = p->ep->Rn * (1-exp(-kappa)) -  Theta*(Tf - p->ep->ta) - Lambda * p->lp->b *
    ( (2.5414e6) * exp( -5415.0 / (Tf + ZeroC)) - p->ep->ea);

  /* update the leaf variables */

  p->lv->cs = p->lp->b;

  p->lv->Ps = (p->ep->Rn * (1-exp(-kappa)) - Theta * (Tf - p->ep->ta))/Lambda;

  p->lv->Vm = p->lp->Vm0 * exp(3000.0 *(1.0/288.2 - 1.0/(Tf + ZeroC)))/
    ((1.0 + exp(0.4 * (5.0 -Tf)))*
     (1.0 + exp(0.4 * (Tf - 100.0))));
 
#if CLOSED_CASE_ASSIMILATION

  /* solve for Ci */

  arubisco = p->lv->Vm;
  Cir1 = 1.6/p->lp->b * (p->lv->Vm * p->lp->littleGamma + p->lp->b * p->ep->ca/1.6-
			 p->lv->Vm);
  p->lv->A = arubisco;

#else

  p->lv->A = 0.0; 		/* gross assimilation is identically 0 */

#endif

  p->lv->A -= p->lv->Vm * p->lp->littleGamma;
  p->lv->Ci = p->ep->ca -  (1.6 * p->lv->A)/p->lv->cs;

  p->lv->tl = Tf;

  return p->lv->A; /* returns the net absorption */

}


/* C4 Plant, A closed, rubsico case 2, Ci limitation */
double AcC4co_2(double Tf, struct pars * p){

  double Ttest;
#if CLOSED_CASE_ASSIMILATION
  double  arubisco;
  double Gamma;			/* CO2 compensation point [mol/mol] */
  double Cir2; 		/* Ci rubisco case (2) */
#endif
  /* note that there are calls to the physical constants Theta, Lamda
     and D0 that are defined globally */

#if DEBU_FUNCTIONS

  printf("AcC4 starting\n");
  printf("\n");

#endif

  /* solve immediately for Tf */
  Ttest = p->ep->Rn * (1-exp(-kappa)) -  Theta*(Tf - p->ep->ta) - Lambda * p->lp->b *
    ( (2.5414e6) * exp( -5415.0 / (Tf + ZeroC)) - p->ep->ea);

  /* update the leaf variables */
	
  p->lv->cs = p->lp->b;

  p->lv->Ps = (p->ep->Rn * (1-exp(-kappa)) - Theta * (Tf - p->ep->ta))/Lambda;

  p->lv->Vm = p->lp->Vm0 * exp(3000.0 *(1.0/288.2 - 1.0/(Tf + ZeroC)))/
    ((1.0 + exp(0.4 * (5.0 -Tf)))*
     (1.0 + exp(0.4 * (Tf - 100.0))));
#if CLOSED_CASE_ASSIMILATION

  /* solve for Ci */

  arubisco = p->lv->Vm;
  Cir2 = ( p->lv->Vm * p->lp->littleGamma + p->lp->b * p->ep->ca/1.6)/(p->lv->Vm * 18000 + p->lp->b/1.6);

  p->lv->Ci =  Cir2;
  arubisco = arubisco * 18000 * p->lv->Ci; /* arubisco is 18000*Vm*Ci */
  p->lv->A = arubisco;

#else

  p->lv->A = 0.0; 		/* gross assimilation is identically 0 */

#endif

  p->lv->A -= p->lv->Vm * p->lp->littleGamma;
  p->lv->Ci = p->ep->ca -  (1.6 * p->lv->A)/p->lv->cs;

  p->lv->tl = Tf;

  return p->lv->A; /* returns the net absorption */

}

/* ********************************************* */
/*                  C3, Open Cases               */
/* ********************************************* */


/* C3 Plant, A open, light case */
double AoC3light(double Tf, struct pars * p){
  double Atest;
  double Gamma;			/* CO2 compensation point [mol/mol] */
  double K1, K2;		/* Parameters for dark reaction */
  double el, Ds;	      /* Leaf H20 concentration and H20 concentration deficit  [mol/mol] */
  /* note that there are calls to the physical constants Theta, Lamda
     and D0 that are defined globally */

#if DEBU_SOLVER

  printf("AoC3 starting with temperature %8.9e\n", Tf);
  printf("\n");

#endif

  p->lv->Ps = ( p->ep->Rn * (1-exp(-kappa)) - Theta * (Tf - p->ep->ta))/Lambda;
	
  el = 2.5414 * 1.0e6 * exp(-5415.0/(Tf+ZeroC));
    
  Ds = el - p->ep->ea;

  p->lv->cs = p->lv->Ps/Ds;

#if DEBU_SOLVER
  printf("Cs calculated at: %8.9e\n", p->lv->cs);
  printf("\n");
#endif

  Gamma = (2.12e-5) * exp(5000.0*(1.0/288.2 - 1.0/(Tf+ZeroC)));

  p->lv->Ci= (p->lp->M * p->lv->cs * p->ep->ca + 1.6 * Gamma * 
	      (1.0 + Ds/D0) * (p->lv->cs - p->lp->b))/(1.6 *(1.0 + Ds/D0)*
						       (p->lv->cs - p->lp->b) + p->lp->M * p->lv->cs);

#if DEBU_SOLVER
  printf("Ci calculated at: %8.9e\n", p->lv->Ci);
  printf("\n");
#endif

  Atest = (p->lv->cs / 1.6 * (p->ep->ca - p->lv->Ci));

#if DEBU_SOLVER
  printf("A(Ci,Cs) calculated at: %8.9e\n", Atest);
  printf("\n");
#endif
	
  K1 = (1.5e-4) * exp(6000.0 *(1.0/288.2 - 1.0/(Tf + ZeroC)));

  K2 = 0.836 * exp(-1400.0 *(1.0/288.2 - 1.0/(Tf + ZeroC)));

  p->lv->Vm = p->lp->Vm0 * exp(3000.0 *(1.0/288.2 - 1.0/(Tf + ZeroC)))/
    (1.0 + exp(0.4 * (5.0 -Tf)))/
    (1.0 + exp(0.4 * (Tf - 45.0)));
	
  p->lv->A = p->lp->alpha * p->ep->L * (1-exp(-kappa)) * (( p->lv->Ci - Gamma)/( p->lv->Ci + Gamma));
		
  if(p->lv->A < 0) p->lv->A = 0.0;

  p->lv->A -= p->lv->Vm * p->lp->littleGamma;


#if DEBU_SOLVER
  printf("A(Tf) calculated at: %8.9e\n", p->lv->A);
  printf("\n");
#endif

  Atest -= p->lv->A;
  p->lv->tl = Tf;


#if DEBU_SOLVER
  printf("The difference between the two A's is: %8.9e\n", Atest);
  printf("\n");
#endif

  return Atest; /* returns the departure from 0 in the solution for Ao */

}


/* C3 Plant, A open, rubisco case */
double AoC3rubisco(double Tf, struct pars * p){
  double Atest;
  double Gamma;			/* CO2 compensation point [mol/mol] */
  double K1, K2;		/* Parameters for dark reaction */
  double el, Ds; /* Leaf H20 concentration and H20 concentration deficit  [mol/mol] */
  /* note that there are calls to the physical constants Theta, Lamda
     and D0 that are defined globally */

#if DEBU_SOLVER

  printf("AoC3 starting with temperature %8.9e\n", Tf);
  printf("\n");

#endif

  p->lv->Ps = ( p->ep->Rn * (1-exp(-kappa)) - Theta * (Tf - p->ep->ta))/Lambda;
	
  el = 2.5414 * 1.0e6 * exp(-5415.0/(Tf+ZeroC));
    
  Ds = el - p->ep->ea;

  p->lv->cs = p->lv->Ps/Ds;

#if DEBU_SOLVER
  printf("Cs calculated at: %8.9e\n", p->lv->cs);
  printf("\n");
#endif

  Gamma = (2.12e-5) * exp(5000.0*(1.0/288.2 - 1.0/(Tf+ZeroC)));

  p->lv->Ci= (p->lp->M * p->lv->cs * p->ep->ca + 1.6 * Gamma * 
	      (1.0 + Ds/D0) * (p->lv->cs - p->lp->b))/(1.6 *(1.0 + Ds/D0)*
						       (p->lv->cs - p->lp->b) + p->lp->M * p->lv->cs);


#if DEBU_SOLVER
  printf("Ci calculated at: %8.9e\n", p->lv->Ci);
  printf("\n");
#endif

  Atest = (p->lv->cs / 1.6 * (p->ep->ca - p->lv->Ci));


#if DEBU_SOLVER
  printf("A(Ci,Cs) calculated at: %8.9e\n", Atest);
  printf("\n");
#endif
	
  K1 = (1.5e-4) * exp(6000.0 *(1.0/288.2 - 1.0/(Tf + ZeroC)));

  K2 = 0.836 * exp(-1400.0 *(1.0/288.2 - 1.0/(Tf + ZeroC)));

  p->lv->Vm = p->lp->Vm0 * exp(3000.0 *(1.0/288.2 - 1.0/(Tf + ZeroC)))/
    (1.0 + exp(0.4 * (5.0 -Tf)))/
    (1.0 + exp(0.4 * (Tf - 45.0)));

  p->lv->A =  p->lv->Vm * ( p->lv->Ci-Gamma)/( p->lv->Ci+K1*(1.0 + K2));
		
  if(p->lv->A < 0) p->lv->A = 0.0;

  p->lv->A -= p->lv->Vm * p->lp->littleGamma;


#if DEBU_SOLVER
  printf("A(Tf) calculated at: %8.9e\n", p->lv->A);
  printf("\n");
#endif

  Atest -= p->lv->A;
  p->lv->tl = Tf;


#if DEBU_SOLVER
  printf("The difference between the two A's is: %8.9e\n", Atest);
  printf("\n");
#endif

  return Atest; /* returns the departure from 0 in the solution for Ao */

}


/* ********************************************* */
/*                  C4, Open Cases               */
/* ********************************************* */


/* C4 Plant, A open, rubisco (no Ci limitation */
double AoC4rubisco(double Tf, struct pars * p){
  double Atest, arubisco, alight;
  double Gamma;			/* CO2 compensation point [mol/mol] */
  double el, Ds;	      /* Leaf H20 concentration and H20 concentration deficit  [mol/mol] */

  /* note that there are calls to the physical constants Theta, Lamda
     and D0 that are defined globally */

  /* reset leaf variables to zero */
  p->lv->tl = Tf;
  p->lv->A = 0.0;
  p->lv->Ps = 0.0;
  p->lv->cs = 0.0;
  p->lv->Ci = 0.0;
  p->lv->Vm = 0.0;

#if DEBU_FUNCTIONS

  printf("AoC4 starting\n");
  printf("\n");

#endif

  p->lv->Ps = ( p->ep->Rn * (1-exp(-0.5)) - Theta * (Tf - p->ep->ta))/Lambda;
	
  el = 2.5414 * 1.0e6 * exp(-5415.0/(Tf+ZeroC));
    
  Ds = el - p->ep->ea;

  p->lv->cs = p->lv->Ps/Ds;

  Gamma = (2.12e-5) * exp(5000.0 *(1.0/288.2 - 1.0/(Tf + ZeroC)));

  p->lv->Ci= (p->lp->M * p->lv->cs * p->ep->ca + 1.6 * Gamma * 
	      (1.0 + Ds/D0) * (p->lv->cs - p->lp->b))/(1.6 *(1.0 + Ds/D0)*
						       (p->lv->cs - p->lp->b) + p->lp->M * p->lv->cs);

#if DEBU_FUNCTIONS

  printf("AoC4: for Tf of %f cs is %f, Ci is %f and Ps is %f\n", Tf,p->lv->cs, p->lv->Ci, p->lv->Ps);

#endif

  Atest = (p->lv->cs / 1.6 * (p->ep->ca - p->lv->Ci));

  p->lv->Vm = p->lp->Vm0 * exp(3000.0 *(1.0/288.2 - 1.0/(Tf + ZeroC)))/
    ((1.0 + exp(0.4 * (5.0 -Tf)))*
     (1.0 + exp(0.4 * (Tf - 100.0))));

  p->lv->A = p->lv->Vm;

  p->lv->A -= p->lv->Vm * p->lp->littleGamma;


#if DEBU_FUNCTIONS

  printf("AoC4: Vm is %f and A is %f\n",
	 p->lv->Vm, p->lv->A);

#endif


  Atest -= p->lv->A;
  p->lv->tl = Tf;

  return Atest; /* returns the departure from 0 in the solution for Ao */

}

/* C4 Plant, A open, Rubisco case, Ci limited */
double AoC4co_2(double Tf, struct pars * p){
  double Atest;
  double Gamma;			/* CO2 compensation point [mol/mol] */
  double el, Ds;	      /* Leaf H20 concentration and H20 concentration deficit  [mol/mol] */

  /* note that there are calls to the physical constants Theta, Lamda
     and D0 that are defined globally */

#if DEBU_FUNCTIONS

  printf("AoC4 starting\n");
  printf("\n");

#endif

  /* reset leaf variables to zero */
  p->lv->tl = Tf;
  p->lv->A = 0.0;
  p->lv->Ps = 0.0;
  p->lv->cs = 0.0;
  p->lv->Ci = 0.0;
  p->lv->Vm = 0.0;

  p->lv->Ps = ( p->ep->Rn * (1-exp(-0.5)) - Theta * (Tf - p->ep->ta))/Lambda;
	
  el = 2.5414 * 1.0e6 * exp(-5415.0/(Tf+ZeroC));
    
  Ds = el - p->ep->ea;

  p->lv->cs = p->lv->Ps/Ds;

  Gamma = (2.12e-5) * exp(5000.0 *(1.0/288.2 - 1.0/(Tf + ZeroC)));

  p->lv->Ci= (p->lp->M * p->lv->cs * p->ep->ca + 1.6 * Gamma * 
	      (1.0 + Ds/D0) * (p->lv->cs - p->lp->b))/(1.6 *(1.0 + Ds/D0)*
						       (p->lv->cs - p->lp->b) + p->lp->M * p->lv->cs);


#if DEBU_FUNCTIONS

  printf("AoC4: for Tf of %f cs is %f, Ci is %f and Ps is %f\n", Tf,p->lv->cs, p->lv->Ci, p->lv->Ps);

#endif


  Atest = (p->lv->cs / 1.6 * (p->ep->ca - p->lv->Ci));

  p->lv->Vm = p->lp->Vm0 * exp(3000.0 *(1.0/288.2 - 1.0/(Tf + ZeroC)))/
    ((1.0 + exp(0.4 * (5.0 -Tf)))*
     (1.0 + exp(0.4 * (Tf - 100.0))));

	
  p->lv->A = p->lv->Vm * 18000 * p->lv->Ci; /* Ci limied rubisco case */
		
  p->lv->A -= p->lv->Vm * p->lp->littleGamma;


#if DEBU_FUNCTIONS

  printf("AoC4 rubsico case 2: Vm is %f and A is %f\n",
	 p->lv->Vm, p->lv->A);

#endif

  Atest -= p->lv->A;
  p->lv->tl = Tf;

  return Atest; /* returns the departure from 0 in the solution for Ao */

}


/* C4 Plant, A open, light limited case - test both sides of the equation */
double AoC4light(double Tf, struct pars * p){


  double Atest;
  double Gamma;			/* CO2 compensation point [mol/mol] */
  double el, Ds;	      /* Leaf H20 concentration and H20 concentration deficit  [mol/mol] */

  /* note that there are calls to the physical constants Theta, Lamda
     and D0 that are defined globally */


#if DEBU_FUNCTIONS

  printf("AoC4 starting\n");
  printf("\n");

#endif

  p->lv->Ps = ( p->ep->Rn * (1-exp(-0.5)) - Theta * (Tf - p->ep->ta))/Lambda;
	
  el = 2.5414 * 1.0e6 * exp(-5415.0/(Tf+ZeroC));
    
  Ds = el - p->ep->ea;

  p->lv->cs = p->lv->Ps/Ds;

  Gamma = (2.12e-5) * exp(5000.0 *(1.0/288.2 - 1.0/(Tf + ZeroC)));

  p->lv->Ci= (p->lp->M * p->lv->cs * p->ep->ca + 1.6 * Gamma * 
	      (1.0 + Ds/D0) * (p->lv->cs - p->lp->b))/(1.6 *(1.0 + Ds/D0)*
						       (p->lv->cs - p->lp->b) + p->lp->M * p->lv->cs);


#if DEBU_FUNCTIONS

  printf("AoC4: for Tf of %f cs is %f, Ci is %f and Ps is %f\n", Tf,p->lv->cs, p->lv->Ci, p->lv->Ps);

#endif


  Atest = (p->lv->cs / 1.6 * (p->ep->ca - p->lv->Ci));

  p->lv->Vm = p->lp->Vm0 * exp(3000.0 *(1.0/288.2 - 1.0/(Tf + ZeroC)))/
    ((1.0 + exp(0.4 * (5.0 -Tf)))*
     (1.0 + exp(0.4 * (Tf - 100.0))));

 	
  p->lv->A = p->lp->alpha * p->ep->L  * (1-exp(-kappa)) ;
		
  if(p->lv->A < 0) p->lv->A = 0.0;

  p->lv->A -= p->lv->Vm * p->lp->littleGamma;


#if DEBU_FUNCTIONS

  printf("AoC4 light case: Vm is %f and A is %f\n",
	 p->lv->Vm, p->lv->A);

#endif


  Atest -= p->lv->A;
  p->lv->tl = Tf;

  return Atest; /* returns the departure from 0 in the solution for Ao */

}



/* copier for the parameters */

int copyPars(struct pars * copyfrom,  struct pars * copyto){

  /* copy environmental parameters */
  copyto->ep->ta =   copyfrom->ep->ta;
  copyto->ep->ca =   copyfrom->ep->ca;
  copyto->ep->ea =   copyfrom->ep->ea;
  copyto->ep->L =   copyfrom->ep->L;
  copyto->ep->Rn =   copyfrom->ep->Rn;

  /* copy leaf variables */
  copyto->lv->A =   copyfrom->lv->A;
  copyto->lv->Ps =   copyfrom->lv->Ps;
  copyto->lv->tl =   copyfrom->lv->tl;
  copyto->lv->Vm =   copyfrom->lv->Vm;
  copyto->lv->cs =   copyfrom->lv->cs;
  copyto->lv->Ci =   copyfrom->lv->Ci;

  /* copy leaf parameters */
  copyto->lp->M =   copyfrom->lp->M;
  copyto->lp->b =   copyfrom->lp->b;
  copyto->lp->alpha =   copyfrom->lp->alpha;
  copyto->lp->littleGamma =   copyfrom->lp->littleGamma;
  copyto->lp->Vm0 =   copyfrom->lp->Vm0;


  return 0;


}


/* **************** */
/* Tester functions */
/* **************** */


/* C4 Plant, A open, rubisco (no Ci limitation */
double AoC4test_rubisco(double Tf, struct pars * p){
  double Atest, arubisco, alight;
  double Gamma;			/* CO2 compensation point [mol/mol] */
  double el, Ds;	      /* Leaf H20 concentration and H20 concentration deficit  [mol/mol] */

  /* note that there are calls to the physical constants Theta, Lamda and D0 that are defined globally */

  double Ps, cs, Ci, A, Vm;



#if DEBU_FUNCTIONS

  printf("AoC4test_ starting\n");
  printf("\n");

#endif

  Ps = ( p->ep->Rn * (1-exp(-0.5)) - Theta * (Tf - p->ep->ta))/Lambda;

  printf("For a Tf of %e Psi is %e (Eq. A12)\n", Tf, Ps);
	
  el = 2.5414 * 1.0e6 * exp(-5415.0/(Tf+ZeroC));
    
  Ds = el - p->ep->ea;

  cs = Ps/Ds;

  printf("Substituting in (A12) in (A10) cs is %e\n", cs);

  Gamma = (2.12e-5) * exp(5000.0 *(1.0/288.2 - 1.0/(Tf + ZeroC)));

  Ci= (p->lp->M * cs * p->ep->ca + 1.6 * Gamma * 
       (1.0 + Ds/D0) * (cs - p->lp->b))/(1.6 *(1.0 + Ds/D0)*
					 (cs - p->lp->b) + p->lp->M * cs);

  printf("Substituting cs in Ci is %e\n", Ci);


  A = (cs / 1.6 * (p->ep->ca - Ci));


  printf("Substituting cs and Ci in A9 A is %e\n", A);


  Vm = p->lp->Vm0 * exp(3000.0 *(1.0/288.2 - 1.0/(Tf + ZeroC)))/
    ((1.0 + exp(0.4 * (5.0 -Tf)))*
     (1.0 + exp(0.4 * (Tf - 100.0))));

  A = Vm;

  A -= Vm * p->lp->littleGamma;

  printf("A from A7 is %e\n", A);

  cs = (p->lp->M * A)/(( Ci - Gamma) * (1.0 + Ds/D0)) + p->lp->b;

  printf("Cs from A8 is %e\n", cs);

  A = cs/1.6 * ( p->ep->ca - Ci);

  printf("A from A9 is %e\n", A);

  Ps = cs * Ds;

  printf("Ps from A10 is %e\n", Ps);

  return Atest; /* returns the departure from 0 in the solution for Ao */

}


/* C4 Plant, A open, Rubisco case, Ci limited */
double AoC4test_co_2(double Tf, struct pars * p){

  double Atest, arubisco, alight;
  double Gamma;			/* CO2 compensation point [mol/mol] */
  double el, Ds;	      /* Leaf H20 concentration and H20 concentration deficit  [mol/mol] */

  /* note that there are calls to the physical constants Theta, Lamda and D0 that are defined globally */

  double Ps, cs, Ci, A, Vm;



#if DEBU_FUNCTIONS

  printf("AoC4test_ starting\n");
  printf("\n");

#endif

  Ps = ( p->ep->Rn * (1-exp(-0.5)) - Theta * (Tf - p->ep->ta))/Lambda;

  printf("For a Tf of %e Psi is %e (Eq. A12)\n", Tf, Ps);
	
  el = 2.5414 * 1.0e6 * exp(-5415.0/(Tf+ZeroC));
    
  Ds = el - p->ep->ea;

  cs = Ps/Ds;

  printf("Substituting in (A12) in (A10) cs is %e\n", cs);

  Gamma = (2.12e-5) * exp(5000.0 *(1.0/288.2 - 1.0/(Tf + ZeroC)));

  Ci= (p->lp->M * cs * p->ep->ca + 1.6 * Gamma * 
       (1.0 + Ds/D0) * (cs - p->lp->b))/(1.6 *(1.0 + Ds/D0)*
					 (cs - p->lp->b) + p->lp->M * cs);

  printf("Substituting cs in Ci is %e\n", Ci);


  A = (cs / 1.6 * (p->ep->ca - Ci));


  printf("Substituting cs and Ci in A9 A is %e\n", A);


  Vm = p->lp->Vm0 * exp(3000.0 *(1.0/288.2 - 1.0/(Tf + ZeroC)))/
    ((1.0 + exp(0.4 * (5.0 -Tf)))*
     (1.0 + exp(0.4 * (Tf - 100.0))));

  A = Vm * 18000 * Ci; /* Ci limied rubisco case */
  A -= Vm * p->lp->littleGamma;
  printf("A from A7 is %e\n", A);

  cs = (p->lp->M * A)/(( Ci - Gamma) * (1.0 + Ds/D0)) + p->lp->b;

  printf("Cs from A8 is %e\n", cs);

  A = cs/1.6 * ( p->ep->ca - Ci);

  printf("A from A9 is %e\n", A);

  Ps = cs * Ds;

  printf("Ps from A10 is %e\n", Ps);

  return Atest; /* returns the departure from 0 in the solution for Ao */

}


/* C4 Plant, A open, light limited case */
double AoC4test_light(double Tf, struct pars * p){
  double Atest, arubisco, alight;
  double Gamma;			/* CO2 compensation point [mol/mol] */
  double el, Ds;	      /* Leaf H20 concentration and H20 concentration deficit  [mol/mol] */

  /* note that there are calls to the physical constants Theta, Lamda and D0 that are defined globally */

  double Ps, cs, Ci, A, Vm;



#if DEBU_FUNCTIONS

  printf("AoC4test_ starting\n");
  printf("\n");

#endif

  Ps = ( p->ep->Rn * (1-exp(-0.5)) - Theta * (Tf - p->ep->ta))/Lambda;

  printf("For a Tf of %e Psi is %e (Eq. A12)\n", Tf, Ps);
	
  el = 2.5414 * 1.0e6 * exp(-5415.0/(Tf+ZeroC));
    
  Ds = el - p->ep->ea;

  cs = Ps/Ds;

  printf("Substituting in (A12) in (A10) cs is %e\n", cs);

  Gamma = (2.12e-5) * exp(5000.0 *(1.0/288.2 - 1.0/(Tf + ZeroC)));

  Ci= (p->lp->M * cs * p->ep->ca + 1.6 * Gamma * 
       (1.0 + Ds/D0) * (cs - p->lp->b))/(1.6 *(1.0 + Ds/D0)*
					 (cs - p->lp->b) + p->lp->M * cs);

  printf("Substituting cs in (A8) Ci is %e\n", Ci);


  A = (cs / 1.6 * (p->ep->ca - Ci));

  


  printf("Substituting cs and Ci in (A9) A is %e\n", A);


  Vm = p->lp->Vm0 * exp(3000.0 *(1.0/288.2 - 1.0/(Tf + ZeroC)))/
    ((1.0 + exp(0.4 * (5.0 -Tf)))*
     (1.0 + exp(0.4 * (Tf - 100.0))));

  A = p->lp->alpha * p->ep->L  * (1-exp(-kappa)) ;
		
  A -= Vm * p->lp->littleGamma;

  printf("A from A7 is %e\n", A);

  cs = (p->lp->M * A)/(( Ci - Gamma) * (1.0 + Ds/D0)) + p->lp->b;

  printf("Cs from A8 is %e\n", cs);

  A = cs/1.6 * ( p->ep->ca - Ci);

  printf("A from A9 is %e\n", A);

  Ps = cs * Ds;

  printf("Ps from A10 is %e\n", Ps);


  return Atest; /* returns the departure from 0 in the solution for Ao */

}
