#ifndef _FARQUAR_H_
#define _FARQUAR_H_

/* solver parameters */
#define SIGN(a,b) ((b) >= 0.0 ? fabs(a) : -fabs(a))
#define NUM_VM 1
#define ITMAX 100
#define EPS 3.0e-16
#define NRANSI
#define MAX_NBR 100000		/* how many tries when fallback is available */
#define TOL 3.0e-8		   /* tolerance for the solver */


/* debugging and output flags */
#define DEBU_WRAPPER 0
#define DEBU_SOLVER 0
#define DEBU_ROOTS 0
#define DEBU_FUNCTIONS 0
#define REPORT_FALLBACK 0


#define CLOSED_CASE_ASSIMILATION 1 /* whether gross photosys is not
				      identically 0 in the closed case
				      (as per the ED Ecol Mon. paper,
				      or it is computed as in the open
				      case but with cs = b */

#define LIGHT_TOLERANCE -10.0 	/* minimum light for the open case attempt */
/* for PAR less than LIGHT_TOLERANCE microEinstain/m2/s use the closed case*/

/* functions to solve the ED paper set of photosynthetic equations given */
/* a set of  arrays of environmental conditions, another set of arrays where to put the  */
/* output, leaf level-parameters and 2 switches (c3/c4 and open/close stomata) */

/*********************************/
/* GLOBALS                       */

/* some phisical constants */
#define Lambda	4.5e-2 /* molar latent heat content of water vapour (J/micromol) */
#define Theta	38.4 /* Governs rate of convective cooling (J/m^2/s/K) */
#define D0      0.01		/* reference H2O mixing ratio */    
#define kappa   0.5	       /* extinction coefficient for leaves */
#define ZeroC 273.15 		/* Zero Celsius in degrees K */

/* physical constatnts to use in the conversion factors */
#define mol2gC   12.01
#define mol2gH2O 18.0153

/** exit conditions for the solver  */
enum solveErr {
  SUCCESS = 0,
  BRACKET_FAILURE,
  CONVERGE_FAILURE,
  FALLBACK,
  FAILURE_OPEN,
  FAILURE_CLOSED
};

/************************************************************************************* */
/*                                   Structures                                        */
/************************************************************************************* */

/* a strcuture to pass all the parameters to the solvers */
struct pars{
 /* pointers  to environmental pars, leaf pars and leaf variables */
  struct envPars * ep;
  struct leafPars * lp;
  struct leafVars * lv;
};

/* a struture to hold all the envinromental parameters - valid for all leaves */
struct envPars{
  double ta;	      /* Atmospheric temperature degrees Celsius */
  double ca;	      /* Atmospheric CO2 concentration mol/mol */
  double ea;          /* ambient humidity [mol/mol]*/
  double L;	      /* Absorbed PAR micro Einsteins m^-2 s^-1 */
  double Rn;	      /* Net radiation J m^-2 s^-1 */

};

/* a strcuture to hold all the leaf-level variables (env. conditions and stomata-opening specific) */
struct leafVars{

  double A; 			/* Photosynthetic rate micromoles C m^-2 s^-1 */
  double Ps;			/* evaporative loss micromoles H2O  m^-2 s^-1 */
  double tl;		       /* Leaf temperatures degrees Celsius */
  double Vm; 		/* Maximum carboxilation capacity  micromol CO2 m^-2 s^-1*/
  double cs;			/* stomatal conductance */
  double Ci;	      /* Leaf internal CO2 concentration mol/mol */
};

/* a structure to hold all the plant type pars (PFT specific, independent form) */
struct leafPars{
/* Transpiration parameters */

  double M;
  double b;			/* slope and intercepts of the relationship between A and cs */

/* Photosysthesys parameters */
  double alpha;			/* Quantum efficiency micromol Einstein^-1 */
  double littleGamma;			/* Leaf respiration proprtion to maximum carboxilation capacity  */
  double Vm0; 		/* Maximum carboxilation capacity  micromol CO2 m^-2 s^-1*/
};


/* a structure to hold the soultions to the open and closed case */

struct solution{
  double Ao;			/* assimilation, open and closed */
  double Ac;
  double Pso;			/* transpiration, open and closed */
  double Psc;
  double cso;			/* stomatal conductancy, open and closed */
  double csc;
  double Cio;			/* intercellular co2, open and closed */
  double Cic;
  double tlo;			/* leaf temperature, open and closed */
  double tlc;
  double Vmo;			/* Vm, open and closed case */
  double Vmc;
};


/********************** Function Prototypes *************************/
/* utilities */
double epsilon(double T, double x, double y); //the dimensionless temperature function
double sign(double x); 		/* return the sign */


/* pass a foliar temprature, return a photosynthetic rate */

double AoC3light(double Tf, struct pars * thepars); 	/* C3 Plant, A open, light case*/
double AoC3rubisco(double Tf, struct pars * thepars); 	/* C3 Plant, A open, rubisco case */

double AoC4light(double Tf, struct pars * thepars); 	/* C4 Plant, A open, light case */
double AoC4rubisco(double Tf, struct pars * thepars); 	/* C4 Plant, A open, rubisco case 1 */
double AoC4co_2(double Tf, struct pars * thepars); 	/* C4 Plant, A open, rubisco case 2 */


/* copier for the parameters */

int copyPars(struct pars * copyfrom,  struct pars * copyto);

/* Closed case, it actually break down in two */

/* solve for Tf from the energy balance of the leaves */
/* this is the function that should be called from the solver */
double TfClosed(double Tf, struct pars * thepars);

/* get A from Tf in the closed cases - these are not to be called from the solver */
/* assume that Tf is the right one  */
double AcC3(double Tf, struct pars * thepars); 	/* A closed C3)*/
double AcC3light(double Tf, struct pars * thepars); 	/* A closed C3 light case */
double AcC3rubisco(double Tf, struct pars * thepars); 	/* A closed C3 rubisco case */

double AcC4(double Tf, struct pars * thepars); 	/* A closed C4 minimum of three cases */
double AcC4light(double Tf, struct pars * thepars); 	/* A closed C4 light case */
double AcC4rubisco(double Tf, struct pars * thepars); 	/* A closed C4 rubisco case  */
double AcC4co_2(double Tf, struct pars * thepars); 	/* A closed C4 co2 case */

/* tester functions */

double AoC4test_light(double Tf, struct pars * p);
double AoC4test_rubisco(double Tf, struct pars * p);
double AoC4test_co_2(double Tf, struct pars * p);

/* solvers */

int MyZbrak(double (*fx)(double, struct pars *thepar), double x1, double x2, int n, double xb1[], double xb2[], int *nb, struct pars * thePars);
double MyZbrent(double (*func)(double, struct pars *), double x1, double x2, double tol, struct pars * thePars);
double myFabs(double arg); 	/* let's see if it is fabs that screws up things */

int SolveFarquar(struct solution * thesol, struct pars * cp, int C3orC4);

#endif /* _FARQUAR_H_ */
