/**
 * @file	mech_lut.c
 * @author	Daniel Lipsitt
 * $Revision: 1.1.2.18 $
 * $Date: 2004/07/26 02:32:48 $
 * 
 * This module has functions for reading and accessing a mechanism
 * lookup table.
 * 
 */

#define _GNU_SOURCE

/* handle files over 2GB (but we can only malloc so much...) */
#ifndef _LARGEFILE_SOURCE
# define _LARGEFILE_SOURCE
#endif
#ifndef _FILE_OFFSET_BITS
# define _FILE_OFFSET_BITS 64
#endif

#include <stdio.h>
#include <stdlib.h>
#include <libiberty.h>
#include <assert.h>
#include <math.h>
#include <string.h>
#include <unistd.h>
#include <sys/mman.h>

#include "mech_lut.h"

/**
 * Looked-up values which are this far from the target will produce a
 * warning */
#define LOOKUP_PCT_TOL 0
/**
 * If non-zero, always report lookup difference. Otherwise only if
 * above LOOKUP_PCT_TOL. */
#define DEBUG_MECH_LOOKUP 0

static struct mech_lut_t {
  float *lut;			/* the lookup table */
  size_t N;			/* length of the table */
  struct param *params[NUMPARAMS];
} mech_lut;

/** 
 * Find the index to the parameter p that gives the value closest to
 * val passed in.
 * 
 * @param p the parameter structure
 * @param target the value to find an index for
 * @param idx (output) an index for p such that p->val[index] is the closest to target 
 * @return difference between target and what we got
 */
double nearest_param_index(struct param *p, float target, size_t *idx) {
  int i = 0;
  float diff, olddiff;
  float percentdiff;
    
  /* this is a slow linear search, but it may be faster than calling log() */
  olddiff = fabs(p->val[i] - target);
  for(i=0; i<p->N; i++) {
    diff = fabs(p->val[i] - target);
    
    if(diff > olddiff)
      break;
    olddiff = diff;
  }
  *idx = i-1;
  
  if( DEBUG_MECH_LOOKUP ) {
    percentdiff = target? olddiff/target*100: 0;
    if(percentdiff >= LOOKUP_PCT_TOL) {
      fprintf(stdout, "%s[%d]\t= %f \t target = %f\t diff = %3.1f%% \n",
	      p->name, i-1, p->val[i-1], target, percentdiff);
    }
  }
  
  return olddiff;
}

/** 
 * Lookup variables corresponding to the given parameters.
 * 
 * @param lut	the lookup array
 * @param params array of param structures
 * @param paramvals an array of values for the parameters
 * @param numparams 
 * @param numvars
 * 
 * @return pointer to an array of variables (ie the data we want to
 * look up.)
 */
static float * const lookup_nearest(float *lut, struct param *params[],
			     float paramvals[], int numparams, int numvars)
{
  int i, j, idx;
  unsigned long offset = 0;
  double diff;			/* sanity check epsilon */
  
  /* add the offsets due to each param index */
  j = 1;
  for (i = numparams-1; i>=0; i--) {
    diff = nearest_param_index(params[i], paramvals[i], &idx);
    offset += j * idx;
    j *= params[i]->N;
  }
  offset *= NMECHVARS;
  assert(offset < mech_lut.N);	/* bounds check */

  return lut + offset;
} /* lookup_nearest() */


/**
 * Params are {CO2 (ppm), Vm0, T (C), % Humidity, net radiation (W/m^2)}
 * 
 * @param paramvals 
 * 
 * @return 
 */
float * const mech_lookup_nearest(float paramvals[]) {
  return lookup_nearest(mech_lut.lut, mech_lut.params, paramvals,
			NUMPARAMS, NMECHVARS);
}

/** 
 * Read the lookup table into a big array.
 * (Uses mmap. Array is read-only)
 * 
 * @param infile file to read from
 * @pre	the file pointer is at the beginning of the file.
 * 
 * @return the lookup table as a big array or NULL on failure
 */
float * const read_lut_array(FILE *infile, struct param *params[],
			     int numparams, int numvars) {
  char *arr = NULL;
  size_t arrsize;		/* number of variables in the table */
  size_t arrbytes;		/* number of bytes in the table */
  int headerbytes;		/* number of bytes in the header */
  int i;
  int fd;			/* file descriptor */
  int filech, arrch;		/* for sanity check */
  fpos_t mark;			/* file position */
  int err;			/* for library call result codes */
  long filebytes;
  int pad = 500;		/* room at end so we can shift array
				   by headerbytes */

  assert(infile);

  arrsize = numvars;
  for(i=0; i<numparams; i++) {
    arrsize *= params[i]->N;
  }
  arrbytes = arrsize * sizeof(float);

  printf("init %'d element mechanism lookup table\n", arrsize);

#warning "is this okay for caller?"
  rewind(infile);

  /* extract the low-level file descriptor from the FILE pointer */
  fd = fileno(infile);
  assert(fd != -1);
  
  /* view as chars at first so we can read the file header */
  arr = (char *) mmap_ro(fd, arrbytes+pad);
  
  headerbytes = header_length(arr);
  assert(headerbytes>0);

  /* advance the pointer past the header */
  arr += headerbytes;
  
  fgetpos(infile, &mark);	/* save our place in file */

  err = fseek(infile, -1, SEEK_END);
  assert(err == 0);

  filech = getc(infile);
  arrch = arr[arrbytes -1];

  //  printf("last chars: file %x arr %x\n", filech, arrch);

#warning "DEBUG"
  err = fseek(infile, -8, SEEK_END);
  assert(err == 0);
  for(i=0; i<8; i++)
    printf("%x ", getc(infile));
  printf("\n");
  for(i=0; i<8; i++)
    printf("%x ", (int) arr[(arrbytes -8 + i)]);
  printf("\n");

  fseek(infile, 0, SEEK_END);
  filebytes = ftell(infile);
  if(filebytes - headerbytes != arrbytes) {
    printf("filesize off by %d bytes\n", (filebytes - headerbytes) - arrbytes);
    printf("file: %d header: %d array: %d\n",
	   filebytes, headerbytes, arrbytes);
    exit(EXIT_FAILURE);
  }

  /* check sanity by comparing last bit of array against file */
  assert(filech == arrch);

  fsetpos(infile, &mark);	/* restore our place in file */
  
  printf("init LUT done.\n");
  
  return (float *) arr;
} /* read_lut_array() */

/** 
 * Read one parameter out of the input file.
 *
 * @pre the file pointer is positioned at the beginning of a parameter
 * line.
 * 
 * @param infile 
 * 
 * @return an initialized  parameter structure or NULL on failure
 */
static struct param * read_param(FILE *infile) {
  struct param *result;
  static const float TOL = 0.001;
  float err;

  /* for getline */
  char *linebuff = NULL;
  size_t bufflen = 0;
  ssize_t len;
  
  /* scanned values */
  char name[256] = "";
  char units[256] = "";
  int N, logscale;
  double min, max, step;

  int count = 0;

  len = getline(&linebuff, &bufflen, infile);
  if(len < 1) return NULL;
  count = sscanf(linebuff, "%s %d %lf %lf %lf %d %s",
		 name, &N, &min, &max, &step, &logscale, units);
  if(count < 6) return NULL;	/* units field is optional */
  result = init_param(name, N, min, step, logscale, units);
  assert(result);
  /* init calculates max. make sure it matches read value. */
  if(abs(result->max - max) / result->max >= TOL) {
    fprintf(stderr, "error: %s: calculated max [%f] differs from file max [%f]"
	    " by more than tolerance\n", name, result->max, max);
    return NULL;
  }
  return result;
} /* read_param() */

/** 
 * Read parameter info from the file header into package variables
 * 
 * @param infile file to read from.
 */
void read_param_header(FILE *infile, struct param *params[NUMPARAMS]) {
  // struct param *CO2, *Vm0, *T, *humidity, *PAR;
  static const int DISCARD = 2;	/* number of header lines before real data */
  
  /* for getline */
  char *linebuff = NULL;
  size_t bufflen = 0;
  ssize_t len;

  int i;
  
  /* throw away the first (c3/c4) line */
  for(i=0; i<DISCARD; i++)
    len = getline(&linebuff, &bufflen, infile);

  for(i = 0; i < NUMPARAMS; i++) {
    params[i] =  read_param(infile);
    assert(params[i]);
  }
  
  if(linebuff)
    free(linebuff);
} /* read_param_header() */

/**
 * Initialize a lookup table with N bins, starting at min and
 * increasing in increments of step.
 */

struct param * init_param(char *name, int N, double min, double step,
		      enum logbool scale, char *units) {
  struct param * l;
  int i;

  /* init memory */
  l = (struct param *) xmalloc(sizeof(struct param));
  l->val = xmalloc(N * sizeof(double));  

  /* init values */
  l->name = strdup(name);
  l->units = strdup(units);
  l->N = N;
  l->min = min;
  l->step = step;
  l->logscale = scale;
  if(l->logscale == LOGSCALE) {
    fill_log_param(l, N, min, step);
  } else {
    for(i = 0; i < N; i++) {
      l->val[i] = min + (step * i);
    }
    l->max = l->val[N-1];
  }
  
  if(DEBUG_MECH_LOOKUP) print_param(stdout, l);
  
  return(l);
  
} /* init_param */

/**
 * Initialise a lookup table with N bins, starting at min and
 * increasing in logarithmic increments of step.
 *
 * @pre Some fields have already been initialized by @see init_param()
 */

void fill_log_param(struct param *l, int N, double min, double step) {
  int i;

  assert(l->name);
  
  /* init values */
  for(i = 0; i < N; i++) {
    l->val[i] = pow(10, min + (i*step)) -1;
  }
  l->max = l->val[N-1];
} /* init_log_param */

void print_param(FILE *outfile, struct param *l) {
  int i;
  
  fprintf(outfile, "LUT %s: %s\n", l->name, l->logscale ? "(LOGSCALE)":"");
  fprintf(outfile, "N:  %d\tstep:\t%f\tmin:\t%f\tmax:\t%f\t\n",
	 l->N, l->step, l->min, l->max);
  for(i = 0; i < l->N; i++)
    fprintf(outfile, " %f\n", l->val[i]);
}

void init_mech_lut(char *infilename) {
  FILE *infile = NULL;
  char dummy[1000];
  int i;
  
  infile = fopen(infilename, "rb");
  if(infile == NULL) {
    fprintf(stderr, "couldn't open lookup table \"%s\"\n", infilename);
    perror(NULL);
    exit(1);
  }

  read_param_header(infile, mech_lut.params);
  fgets(dummy, 1000, infile); /* discard line between header and data */
  mech_lut.lut = read_lut_array(infile, mech_lut.params, NUMPARAMS, NMECHVARS);
  assert(mech_lut.lut != NULL);

  /* calculate the size of the LUT */
  mech_lut.N = NMECHVARS;
  for(i=0; i<NUMPARAMS; i++) {
    mech_lut.N *= mech_lut.params[i]->N;
  }

  printf("read lut successfully.\n");

  return;
}

/** 
 * mmap a file readonly or die.
 * 
 * @param fd file
 * @param length number of bytes to map
 * 
 * @return pointer to map
 */
void * const mmap_ro(int fd, size_t length) {
  void * result;
  
  result = mmap(0, length, PROT_READ, MAP_PRIVATE | MAP_NORESERVE, fd, 0);

  if (result == MAP_FAILED) {
    perror(NULL);
    exit(EXIT_FAILURE);
  }
  
  return result;
}

/** 
 * Takes a pointer to an array with no null-termination and tries to
 * find the location after the header ends. Die on failure.
 * 
 * @param arr 
 * 
 * @return pointer to first char in arr after end of header
 */
char * remove_header(const char *arr) {
  char * loc = NULL;
  size_t delimlen;		/* delimiter length */
  
  delimlen = strlen(HEADER_END_DELIM);
  
  /* we subtract -1 from delimlen b/c we don't want the null */
  loc = memmem(arr, MAX_HEADER_LEN, HEADER_END_DELIM,
		       delimlen -1);
  if(NULL == loc) {
    fprintf(stderr, "Couldn't find end of LUT header. Possible malformed LUT");
    exit(1);
  }
  /* loc now points to the 1st char of the HEADER_END_DELIM. next
     advance the pointer past the delim string. */
  loc += delimlen;

#if 0
#warning "debug code for header removal"
  printf("last 3 chars of header: '%c%c%c'\n", loc[-3], loc[-2], loc[-1]);
  printf("first 3 bytes of data: 0x%x 0x%x 0x%x\n", loc[0], loc[1], loc[2]);
#endif

  return loc;  
}


/** 
 * Takes a pointer to an array with no null-termination and tries to
 * find the length of the header.
 * 
 * @param arr 
 * 
 * @return length of the header
 */
int header_length(const char *arr) {
  char * loc = NULL;
  size_t delimlen;		/* delimiter length */
  
  delimlen = strlen(HEADER_END_DELIM);
  
  /* we subtract -1 from delimlen b/c we don't want the null */
  loc = memmem(arr, MAX_HEADER_LEN, HEADER_END_DELIM,
		       delimlen -1);
  if(NULL == loc) {
    fprintf(stderr, "Couldn't find end of LUT header. Possible malformed LUT");
    exit(1);
  }
  /* loc now points to the 1st char of the HEADER_END_DELIM. next
     advance the pointer past the delim string. */
  loc += delimlen;

  return loc - arr;
}
