/* test solvers directly */

#include "farquar.h"
#include <getopt.h>
#include <stdlib.h>
#include <string.h>
#include <libiberty.h>
#include <assert.h>
#include <float.h>		/* provides FLT_EPSILON */

#include "mech_lut.h"
#include "EDMech.h"

/* commandline option flags */
#define CO2_DEFAULT 367		/* default concentration in ppm */

/** PAR above which open case calculations do not fail. */
static const double PAR_EPSILON = 3.0;

enum Cx {
  C3 = 0,
  C4 = 1
};

static int solve(struct envPars *theEnv, struct solution * thesol,
		 struct pars * cp, int C3orC4);

int main(int argc, char *argv[])
{

  struct envPars theEnv;
  struct leafPars c3lf, c4lf;	/* c3 and c4 leaves */
  struct leafVars cl;		/* current leaf */
  struct pars cp;		/* current pars */

  /* solver related variables */
  float paramvals[5];
  
  struct solution soln;		/* Farquar solution output*/

  int i;

  if(argc != 6) {
    fprintf(stderr, "usage: CO2 (ppm), Vm0, temp (C), %% humidity, "
	    "net radiation (W/m^2) \n");
    exit(1);
  }

  for (i=0; i<5; i++) {
    paramvals[i] = atof(argv[i+1]);
  }

  /* Assign initial values to leaf parameters */
  /* M, b, alpha, littleGamma, Vm0 */

  c3lf.M = 8.0;
  c3lf.b = 1.0e4;
  c3lf.alpha = 0.08;
  c3lf.littleGamma = 0.02;

  c4lf.M = 4.0;
  c4lf.b = 0.01 * 1.0e6;
  c4lf.alpha = 0.06;
  c4lf.littleGamma = 0.04;

  /* set the pointers */
  cp.lv = &cl;
  cp.ep = &theEnv;

  /* set leaf params to c3 or c4*/
  cp.lp = &c3lf;
  theEnv.ca = paramvals[0] * 1.0e-6;
  c3lf.Vm0 = c4lf.Vm0 = paramvals[1];
  theEnv.ta = paramvals[2];
    
  /* convert humidity to mol/mol */
    
  theEnv.ea = convert_humidity(paramvals[3], theEnv.ta);

  theEnv.L = paramvals[4] * convW2mkEins * convRn2PAR;

  /* convert PAR in J*m^-2*s^-1 to Net Shortwave Radiation
     in J*m^-2*s^-1 */
  theEnv.Rn = paramvals[4]; //theEnv.L * convPAR2Rn/convW2mkEins;
  printf("Rn = %f\n",theEnv.Rn);

  printf("C3:\n");
  solve(&theEnv, &soln, &cp, C3);
  
  /* reset all values to zero  */
  cl.A = 0.0;
  cl.Ps = 0.0;
  cl.tl = 0.0;
  cl.cs = 0.0;
  cl.Ci = 0.0;

  /* set leaf params to c4*/
  cp.lp = &c4lf;

  printf("\nC4:\n");
  solve(&theEnv, &soln, &cp, C4);

  return 0;
} /* end of main function */


/**
 * calculate and print the solution
 */
static int solve(struct envPars *theEnv, struct solution * soln,
		  struct pars * cp, int C3orC4)
{
  int status;			/* return flag for solver */
  
  status =  SolveFarquar(soln, cp, C3orC4);
	      
  switch (status) {
  case SUCCESS:
    /* we're fine */
    break;
  case FALLBACK:
    /* solver fell back to closed case. report it? */
    if(REPORT_FALLBACK) {
      printf("solver fell back to closed case "
	     " (%e PAR %e RNET %e Ea %e Tf)"
	     " Falling back to the closed case.\n",
	     theEnv->L, theEnv->Rn, theEnv->ea, theEnv->ta);
    }
    break;
  default:
    /* solver failed unrecoverably. return error conditon. */
    printf("Solver failed with status code %d:\n", status );
    print_envPars(stdout, cp->ep);
    return 1;
  }

  printf("month hour Vm0 PAR Rn Ta ea ca Tf cs ci Ao Ac Pso Psc \n");
  printf("%d %d %.2f %f %f %f %f"
	 " %f %f %f %f %f %f %f %f\n",
	 0, 0, cp->lp->Vm0, theEnv->L, theEnv->Rn, theEnv->ta, theEnv->ea,
	 theEnv->ca, cp->lv->tl, cp->lv->cs, cp->lv->Ci, soln->Ao, soln->Ac,
	 soln->Pso, soln->Psc );	      
  //print_result(soln->Ao, soln->Pso, soln->Ac, soln->Psc);
  printf("%e\t%e\t%e\t%e\t%e\t%e\n", soln->Ao, soln->Ac, soln->Pso, soln->Psc,
	 soln->tlo, soln->tlc);
  return 0;
}

/** 
 * Try to open a file with the given mode. Exit if open fails.
 */
FILE * open_or_die(char* filename, char* mode) {
  FILE * file = NULL;
  file = fopen(filename, mode);
  if (file == NULL) {
    printf("Cannot open %s\n", filename);
    exit (EXIT_FAILURE);
  }
  return file;
}

#if 0
/**
 * Convert islscp coordinates to ED and trim file extension
 */
static char * convert_islscp_filename(char * name) {
  float lat, lon;
  char * result;
  result = malloc(256*sizeof(char));
  
  name = basename(name);
  if(sscanf(name, "lat%flon%f.pin", &lat, &lon) != 2) {
    fprintf(stderr, "couldn't scan input filename: %s\n", name);
    exit(EXIT_FAILURE);
  }
  /* change coords from NW corner to center */
  lat -= 0.5;
  lon += 0.5;
  /* construct the result */
  snprintf(result, 256, "lat%.1flon%.1f", lat, lon );
  return result;
}
#endif

/** 
 * Prints args to binary and/or text output file.
 * 
 * @param Ao 
 * @param Pso 
 * @param Ac 
 * @param Psc 
 */
void print_result(double Ao, double Pso, double Ac, double Psc) {
  printf("%e\t%e\t%e\t%e\n", Ao, Ac, Pso, Psc);
}

/**
 * Convert a % humidity to a mixing ratio in
 * mol/mol at a given centigrade temperature.
 */
double convert_humidity(double h, double t) {
  t += ZeroC;			// convert to kelvin
  return(exp(-5415.0/t) * 2.5414e6 * h/100.0);
}

void print_envPars(FILE *outfile, struct envPars *e) {
  fprintf(outfile, "envPars:\n");
  fprintf(outfile, "    T: %f (deg C)\n", e->ta);
  fprintf(outfile, "  CO2: %f (mol/mol)\n", e->ca);
  fprintf(outfile, " hum.: %f (mol/mol)\n", e->ea);
  fprintf(outfile, " absorbed PAR: %f (uEinsteins m^-2 s^-1)\n", e->L);
  fprintf(outfile, " net rad.: %f (J m^-2 s^-1)\n", e->Rn);
}
